import os
import sqlite3
import json
import yaml
import pandas as pd
import numpy as np
from tqdm import tqdm
from glob import glob
from datetime import datetime, timedelta
from termcolor import colored, cprint
import traceback
from tabulate import tabulate
from korus.util import collect_audiofile_metadata
import korus.db as kdb
from korus.util import list_to_str
import korus.app.app_util.view as vw
import korus.app.app_util.ui as ui


def save_changes_to_db(conn):
    """ Helper function for saving changes to the database"""
    try:
        save = ui.UserInputYesNo(
            "save_changes", 
            "Save changes (if any) to the local Korus database? [y/N]", 
        ).request()

    except KeyboardInterrupt:
        terminate(conn, save=False)

    if save:
        conn.commit()
        cprint(f" ## Successfully updated local Korus database", "yellow")


def terminate(conn, save=True):
    """ Helper function for gracefully terminating the program"""
    cprint("\n ## Terminating ...", "yellow")
    if save:
        save_changes_to_db(conn)
    
    cprint("\n ## Closing database connection and exiting ...", "yellow")
    conn.close()
    exit(1)


def add_deployment(conn, logger):
    """ Interactive session for adding a new deployment to the database.

        Args:
            conn: sqlite3.Connection
                Database connection
            logger: korus.app.app_util.ui.InputLogger
                Input logger

        Returns:
            : int
                Row index of the new entry
    """
    cprint(f"\n ## Collecting metadata about the deployment", "yellow")

    # query user for input
    owner = ui.UserInput("owner", "Data owner/collector", group="deployment").request(logger)
    name = ui.UserInput("name", "Descriptive name for the deployment", group="deployment").request(logger)
    start_utc = ui.UserInput("start_utc", "UTC start date or time (e.g. 2020-06-01)", group="deployment").request(logger) 
    end_utc = ui.UserInput("end_utc", "UTC end date or time (e.g. 2020-06-01)", group="deployment").request(logger) 
    location = ui.UserInput("location", "Geographic location (e.g. Salish Sea, BC, Canada)", group="deployment").request(logger)
    latitude_deg = ui.UserInput("latitude_deg", "Latitude (degrees)", transform_fcn=float, group="deployment", unknown=True).request(logger)
    longitude_deg = ui.UserInput("longitude_deg", "Longitude (degrees)", transform_fcn=float, group="deployment", unknown=True).request(logger)
    depth_m = ui.UserInput("depth_m", "Depth (meters)", transform_fcn=float, group="deployment", unknown=True).request(logger)
    sample_rate = ui.UserInput("sample_rate", "Sampling rate in samples/s per channel", transform_fcn=int, group="deployment").request(logger)
    data_license = ui.UserInput("data_license", "License/terms for access to and use of data", group="deployment", unknown=True).request(logger)
    hydrophone = ui.UserInput("hydrophone", "Hydrophone make and model", group="deployment").request(logger)
    bits_per_sample = ui.UserInput("bits_per_sample", "Bits per sample", transform_fcn=int, group="deployment").request(logger)
    num_channels = ui.UserInput("num_channels", "Number of recording channels", transform_fcn=int, group="deployment").request(logger)
    sensitivity = ui.UserInput("sensitivity", "Hydrophone sensitivitiy in dB re 1 V/muPa", transform_fcn=float, group="deployment", unknown=True).request(logger)
    comments = ui.UserInput("comments", "Any additional observations", group="deployment", unknown=True).request(logger)

    # collect the data in a dictionary
    data = {
        "owner": owner,  # owner of the data, can be a person or an organisation
        "name": name,  # descriptive name for the deployment
        "start_utc": start_utc, # UTC start time of recording
        "end_utc": end_utc, # UTC end time of recording
        "location": location, # descriptive name for the deployment location
        "latitude_deg": latitude_deg, # latitude, in degrees
        "longitude_deg": longitude_deg, # longitude, in degrees
        "depth_m": depth_m, # depth, in meters
        "license": data_license, # license/terms governing access to and use of data
        "hydrophone": hydrophone, # hydrophone make and model
        "bits_per_sample": bits_per_sample, # bits per sample
        "sample_rate": sample_rate, # sampling rate in samples/s
        "num_channels": num_channels, # no. channels
        "sensitivity": sensitivity, # calibrated sensitivity in ...
        "comments": comments # additional comments
    }

    # print parsed user input
    cprint("\n ## User input:", "yellow")
    for k,v in data.items():
        cprint(f"    * {k}: {v}", "yellow")

    # submit it to the database
    cursor = kdb.insert_row(conn, table_name="deployment", values=data)

    # commit changes
    #conn.commit()

    cprint(f"\n ## Successfully added deployment `{name}` to the database", "yellow")

    return cursor.lastrowid


def add_job(conn, logger):
    """ Interactive session for adding a new annotation job to the database.

        TODO: remove background_sound

        Args:
            conn: sqlite3.Connection
                Database connection
            logger: korus.app.app_util.ui.InputLogger
                Input logger

        Returns:
            : int
                Row index of the new entry
    """
    cprint(f"\n ## Collecting metadata about the annotation job", "yellow")

    default_prim = [
        ("Cetacean","TC"),
        ("KW","PC"),
        ("KW","W"),
    ]

    # query user for input

    ui_taxonomy_id = ui.UserInput("taxonomy_id", "Taxonomy ID (1,2,3,...)", group="job")
    ui_taxonomy_id.add_option(
        key=["v","view"],
        message="View existing taxonomies",
        fcn=lambda x: vw.view_taxonomies(conn)
    )
    taxonomy_id = ui_taxonomy_id.request(logger)

    annotator = ui.UserInput("annotator", "Annotator (initials only)", group="job").request(logger)
    start_utc = ui.UserInput("start_utc", "UTC start date (e.g. 2020-06-01)", group="job").request(logger) 
    end_utc = ui.UserInput("end_utc", "UTC end date (e.g. 2020-06-01)", group="job").request(logger)
    is_exhaustive = ui.UserInputYesNo("is_exhaustive", "Were all primary sounds annotated? [y/N]", group="job").request(logger)

    if is_exhaustive:
        primary_sound = ui.UserInputSound(
            name="primary_sound", 
            message='Primary sounds that were systematically annotated, e.g. [(\'HW\',\'Upsweep\'), (\'HW\',\'Moan\')]',
            conn=conn,
            taxonomy_id=taxonomy_id,
            group="job",
            default=default_prim,
        ).request(logger)
    
    else:
        primary_sound = None

    """
    background_sound = ui.UserInputSound(
        name="background_sound", 
        message='Background sounds that were only opportunistically annotated, e.g., (\'Anthro\',\'%\')',
        conn=conn,
        taxonomy_id=taxonomy_id,
        group="job",
        default=default_backgr,
    ).request(logger)
    """
    background_sound = None

    # prompt user for issues
    ui_add_another_issue = ui.UserInputYesNo(
        "add_another_issue", 
        "Do the annotations have any known issues? [y/N]", 
    )

    issues = []
    while True:
        if ui_add_another_issue.request():
            description = ui.UserInput("add_issue", "Briefly describe the issue").request(logger)
            issues.append(description)

            ui_add_another_issue.message = "Do the annotations have any other known issues? [y/N]"
        else:
            break


    def fcn(x):
        """Helper function for transforming user input for `comments`"""
        # user provides the path to a plain text file with the comments
        if os.path.exists(x):
            with open(x, "r") as f:
                return f.read()

        # user types in the comments in the console
        else:
            return x

    comments = ui.UserInput(
        "comments", 
        "Any additional observations (type comments in console, or provide path to a plain text file)", 
        group="job",
        transform_fcn=fcn,
    ).request(logger)

    # collect in dict
    data = {
        "taxonomy_id": taxonomy_id,
        "annotator": annotator,
        "is_exhaustive": is_exhaustive,
        "start_utc": start_utc,
        "end_utc": end_utc,
        "issues": json.dumps(issues),
        "comments": comments
    }
    if primary_sound is not None and primary_sound != "":
        data["primary_sound"] = primary_sound
    if background_sound is not None and background_sound != "":
        data["background_sound"] = background_sound

    # print parsed user input
    cprint("\n ## User input:", "yellow")
    for k,v in data.items():
        cprint(f"    * {k}: {v}", "yellow")

    # insert into the database
    cursor = kdb.insert_job(conn, values=data)

    # commit changes
    #conn.commit()

    cprint(f"\n ## Successfully added annotation job to the database", "yellow")

    return cursor.lastrowid


def add_data_storage_location(conn, logger):
    """ Interactive session for adding a new data storage location to the database.

        Args:
            conn: sqlite3.Connection
                Database connection
            logger: korus.app.app_util.ui.InputLogger
                Input logger

        Returns:
            : int
                Row index of the new entry
    """
    cprint(f"\n ## Collecting metadata about the data storage location", "yellow")

    def validate_path(x):
        if not os.path.exists(x):
            raise FileNotFoundError
        return x

    name = ui.UserInput(
        "name", 
        "Descriptive name for the data storage location", 
        group="storage",
    ).request(logger)

    audio_path = ui.UserInput(
        "path", 
        "Full path to the data folder",
        group="storage",
        transform_fcn=validate_path,
    ).request(logger)

    description = ui.UserInput(
        "description", 
        "Provide a highly condensed description of the contents of the data folder",
        group="storage",
    ).request(logger)

    # insert into the database
    v = {
        "name": name,
        "path": audio_path,
        "description": description,        
    }
    cursor = kdb.insert_row(conn, table_name="storage", values=v)

    # commit changes
    #conn.commit()

    cprint(f"\n ## Successfully added data storage location to the database", "yellow")

    return cursor.lastrowid


def create_timestamp_parser(group=None, logger=None):
    """ Interative session for creating a timestamp parser 
    
        Args:
            group: str
                Group that the parameter belongs to. Optional. 
                Parameter names must be unique within groups.
            logger: korus.app.app_util.ui.InputLogger
                Input logger

        Returns:
            timestamp_parser: callable
                Takes a string as input and returns a datetime object
    """
    ui_ts_fmt = ui.UserInput(
        "timestamp_format", 
        "Timestamp format, e.g., %Y%m%dT%H%M%S.%f", 
        group=group,
    )
    ui_ts_fmt.add_option(
        key=["s","smru"],
        message="SMRU timestamp format",
        fcn=lambda x: "%Y%m%d_%H%M%S_%f"
    )    
    ts_fmt = ui_ts_fmt.request(logger) 

    ts_offset = ui.UserInput(
        "timestamp_offset", 
        "Time zone offset (hours) relative to UTC", 
        group=group,
        transform_fcn=float,
    ).request(logger)

    ui_ts_siz = ui.UserInput(
        "timestamp_size", 
        "Timestamp length (no. characters)", 
        group=group,
        transform_fcn=int,
    )
    ui_ts_siz.add_option(
        key=["s","smru"],
        message="SMRU timestamp format",
        fcn=lambda x: 19
    )
    ts_siz = ui_ts_siz.request(logger)

    ts_pos = ui.UserInput(
        "timestamp_position", 
        "Timestamp position (no. characters from left if positive / from right if negative)", 
        group=group,
        transform_fcn=int,
    ).request(logger)

    def timestamp_parser(x):        
        fname = os.path.basename(x)
        p = ts_pos if ts_pos >= 0 else len(fname) + ts_pos - ts_siz
        dt_str = fname[p : p + ts_siz]
        dt = datetime.strptime(dt_str, ts_fmt)
        dt -= timedelta(seconds=int(ts_offset*3600))
        return dt
    
    return timestamp_parser


def add_files(conn, deployment_id, start_utc, end_utc, logger):
    """ Interactive session for adding a new audio files to the database.

        Args:
            conn: sqlite3.Connection
                Database connection
            deployment_id: int
                Deployment index
            start_utc, end_utc: datetime.datetime
                UTC time range
            logger: korus.app.app_util.ui.InputLogger
                Input logger

        Returns:
            : int
                Row index of the last added audio file
    """

    # query user for data-storage location

    def transform_fcn(x):
        id = int(x)
        c = conn.cursor()
        ids = [row[0] for row in c.execute("SELECT id FROM storage").fetchall()]
        if id not in ids:
            raise ValueError(f"Database does not contain data storage location with ID {id}")
        
        return id

    x = ui.UserInput(
        group="storage",
        name="id",
        message="Enter data storage location ID (1,2,3,...) or add a new data storage location",
        transform_fcn=transform_fcn,
    )
    x.add_option(
        key=["v","view"],
        message="View existing data storage locations",
        fcn=lambda x: vw.view_data_storage_locations(conn)
    )
    x.add_option(
        key=["n","new"],
        message="New data storage location",
        fcn=lambda x: add_data_storage_location(conn, logger)
    )

    try:
        storage_id = x.request(logger)

    except KeyboardInterrupt:
        terminate(conn)


    # query user for further information about the audio files

    (audio_path,) = conn.cursor().execute(f"SELECT path FROM storage WHERE id = {storage_id}").fetchall()[0]

    audio_format = ui.UserInput(
        "audio_format", 
        "Audio file format", 
        group="audio",
    ).request(logger)

    # timestamp parser
    timestamp_parser = create_timestamp_parser("audio", logger)

    date_subfolder = ui.UserInputYesNo(
        "date_subfolder", 
        "Are audio files organized in date-stamped subfolders named `yyymmdd`? [y/N]", 
    ).request(logger)

    # automatically search for audio files and collect metadata

    cprint(f" ## Scanning for {audio_format} files ... (this may take a while)", "yellow")

    # first, only obtain the timestamps (fast)
    df = collect_audiofile_metadata(
        path=audio_path, 
        ext=audio_format, 
        timestamp_parser=timestamp_parser,
        earliest_start_utc=start_utc,
        latest_start_utc=end_utc,
        progress_bar=True,
        date_subfolder=date_subfolder,
        inspect_files=False,
    )

    cprint(f" ## Found {len(df)} {audio_format} files in the folder {audio_path} between {start_utc} and {end_utc}", "yellow")

    if len(df) == 0:
        return 0, timestamp_parser

    def fcn(x):
        """Helper function for transforming user input for select_files option"""
        # user provides the path to a plain text file with one filename per row
        if os.path.exists(x):
            with open(x, "r") as f:
                lines = f.readlines()
                return [os.path.basename(line.strip()) for line in lines]

        # user provides a list of file IDs or filenames
        else:
            return [y if "." in y else int(y) for y in x.split(",")]
        
    ui_select_files = ui.UserInput(
        "select_files", 
        "Select the files that you want to add to the database, e.g., 0,3,16 or file1.flac,file2.flac or filenames.txt", 
        transform_fcn=fcn,
        group="audio",
    )

    def view_files_fcn():
        for i,(idx,row) in enumerate(df.iterrows()): 
            print(i, row.to_dict())

    ui_select_files.add_option(
        key=["v","view"],
        message="View files",
        fcn=lambda x: view_files_fcn(),
    )

    ui_select_files.add_option(
        key=["a","all"],
        message="Select all files",
        fcn=lambda x: [i for i in range(df.shape[0])],
    )

    file_ids = ui_select_files.request()

    if isinstance(file_ids[0], int):
        df = df.iloc[file_ids]
    else:
        df = df[df.filename.isin(file_ids)]

    # now, inspect the selected files (slow) to also obtain file size and sampling rate
    rel_path = df.apply(lambda r: os.path.join(r.relative_path, r.filename), axis=1)
    df = collect_audiofile_metadata(
        path=audio_path, 
        ext=audio_format, 
        timestamp_parser=timestamp_parser,
        subset=rel_path.values,
        progress_bar=True,
    )

    # check for table format to ensure backward compatibility with Korus versions <= 0.0.2
    c = conn.cursor()
    columns = [i[1] for i in c.execute("PRAGMA table_info(file)")]
    old_fmt = "dir_path" in columns

    # insert files into database, one by one
    file_counter = 0
    for _,row in df.iterrows(): 
        data = row.to_dict() #convert pd.Series to dict

        # additional required metadata
        data["storage_id"] = storage_id
        data["deployment_id"] = deployment_id

        # to ensure backward compatibility with Korus versions <= 0.0.2
        if old_fmt:
            data["dir_path"] = data["relative_path"]
            data["location"] = audio_path

        # insert in the 'file' table
        try:
            cursor = kdb.insert_row(conn, table_name="file", values=data)
            file_counter += 1

        except sqlite3.IntegrityError:
            # skip if file already exists in db
            fname = data["filename"]
            cprint(f" ## File {fname} already in database, skipping ...", "red")            

    # commit changes
    #conn.commit()

    cprint(f"\n ## Successfully added {file_counter} audio files to the database", "yellow")

    return cursor.lastrowid, timestamp_parser


def add_annotations(conn, deployment_id, job_id, logger, timestamp_parser=None):
    """ Interactive session for adding new annotations to the database.

        Args:
            conn: sqlite3.Connection
                Database connection
            deployment_id: int, list(int)
                Deployment index or indices
            job_id: int
                Annotation job index
            logger: korus.app.app_util.ui.InputLogger
                Input logger
            timestamp_parser: callable
                Given a filename (str) returns the UTC timestamp embedded 
                in the filename as a datetime object.
                Only required if some of the annotations reference audio files not present in the database.

        Returns:
            None
    """
    if isinstance(deployment_id, int):
        deployment_id = [deployment_id]

    # query user
    path = ui.UserInput(
        "path", 
        "Full path to RavenPro selection table. (Use comma to separate multiple files. Use of wildcards is also supported.)", 
        group="annotation",
        transform_fcn=lambda x: x.split(",") if "," in x else x
    ).request(logger)

    # get all file IDs
    c = conn.cursor()
    query = f"SELECT id,deployment_id,filename FROM file WHERE deployment_id IN {list_to_str(deployment_id)}"
    rows = c.execute(query).fetchall()
    
    # check that filenames are unique across deployments
    filenames = [row[2] for row in rows]
    if len(filenames) > len(set(filenames)):
        cprint(" ## Non-unique filename across deployments", "red")
        terminate(conn)

    file_id_map = {row[2]: (row[0], row[1]) for row in rows}  # filename -> (file ID, deployment ID) mapping

    # load taxonomy
    c = conn.cursor()
    tax_id = c.execute(f"SELECT taxonomy_id FROM job WHERE id = '{job_id}'").fetchall()[0][0]
    tax = kdb.get_taxonomy(conn, tax_id)

    annot_ids = []
    while True:
        try:
            # prompt user for granularity
            query = "SELECT name FROM granularity"
            rows = c.execute(query).fetchall()
            allowed_values = list(np.unique([row[0] for row in rows]))
            granularity = ui.UserInput(
                name="granularity",
                message=f"Annotation granularity.",
                group="annotation",
                allowed_values=allowed_values,
            ).request(logger)

            # load selection table(s)
            df = from_raven(
                path, 
                tax, 
                timestamp_parser=timestamp_parser, 
                granularity=granularity, 
                interactive=True,
                progress_bar=True,
            )

            # add missing fields
            # (set file ID to zero (0) if file is missing)
            df["job_id"] = job_id
            df["file_id"] = df.path.apply(lambda x: file_id_map.get(os.path.basename(x), [0])[0])
            if len(deployment_id) == 1:
                df["deployment_id"] = deployment_id[0]
            else:
                df["deployment_id"] = df.path.apply(lambda x: file_id_map[os.path.basename(x)][1])

            # drop columns
            df = df.drop(columns=["path"])

            # if the selection table contains any tags, check if they need to be added to the database
            if "tag" in df.columns.values:
                tag_count = unique_from_list(df.tag)
                tags = list(tag_count.keys())
                new_tags = []
                for tag in tags:
                    tag_id = c.execute(f"SELECT id FROM tag WHERE name = '{tag}'").fetchone()
                    if tag_id is None:
                        new_tags.append(tag)

                if len(new_tags) > 0:
                    cprint(f"\n ## The selection table contains the following new tags: {new_tags}", "red")

                    proceed = ui.UserInputYesNo(
                        "add_tags", 
                        "Add the new tags to the database? [y/N]  (required to proceed with submission)", 
                    ).request()

                    if not proceed:
                        terminate(conn)

                    add_tags(conn, tags)
                    
            # insert into db
            cprint(f"\n ## Adding {len(df)} annotations to the database ...", "yellow")

            start = datetime.now()
            annot_ids = kdb.add_annotations(conn, annot_tbl=df, job_id=job_id, progress_bar=True)
            end = datetime.now()

            break

        except KeyboardInterrupt:
            terminate(conn)

        except Exception:
            cprint("\n ## "+ traceback.format_exc(), "red")
            cprint(" ## Error processing selection table", "red")
            terminate(conn)


    cprint(f"\n ## Successfully added {len(annot_ids)} annotations to the database in {(end - start).total_seconds():.2f} seconds", "yellow")

    # check if any of the annotations just added pertain to audio files not present in the database
    c = conn.cursor()
    query = f"SELECT file_id FROM annotation WHERE id in {list_to_str(annot_ids)}"
    rows = c.execute(query).fetchall()
    num_missing = 0
    for row in rows:
        num_missing += (row[0] == 0)

    if num_missing > 0:
        cprint(f"\n ## WARNING: {num_missing} of the annotations pertain to audio files not present in the database", "red")

    # print summary
    if len(annot_ids) > 0:
        print_summary = ui.UserInputYesNo(
            "print_summary", 
            "Print summary? [y/N]", 
            group="annotation"
        ).request()

        if print_summary:  
            print_annotation_summary(conn, annot_ids)


def unique_from_list(x):
    """ Helper function for extracting unique values
        from a pandas DataFrame column which contains list objects 

        Args:
            x: pandas.Series
                The column containing the list objects
        Returns:
            value_count: dict
                Unique values and the number of times they each occur, 
                sorted in order of decreasing occurrence frequency.
    """
    value_count = {}
    for idx, obj in x[~x.isna()].items():
        for val in obj:
            value_count[val] = value_count.get(val, 0) + 1

    value_count = dict(sorted(value_count.items(), key=lambda item: item[1], reverse=True))
    return value_count


def add_tags(conn, tags):
    """ Interactive session for adding new annotations to the database.

        Args:
            conn: sqlite3.Connection
                Database connection
            tags: str, list(str)
                Tag name(s)

        Returns:
            None
    """
    if not isinstance(tags, list):
        tags = [tags]

    for tag_name in tags:
        cprint(f"\n ## Adding the tag `{tag_name}` to the database ...", "yellow")

        x = ui.UserInput(
            group="annotation",
            name="tag_description",
            message="Tag description",
        )
        x.add_option(
            key=["v","view"],
            message="View existing tags",
            fcn=lambda x: vw.view_tags(conn)
        )

        tag_description = x.request()

        v = {
            "name": tag_name, 
            "description": tag_description,
        }
        kdb.insert_row(conn, table_name="tag", values=v)

        cprint(f"\n ## Successfully added the tag `{tag_name}` to the database", "yellow")


def from_raven(input_path, tax, granularity, sep=None, timestamp_parser=None, interactive=True, progress_bar=False):
    """ Loads entries from a RavenPro selections table

        Args:
            input_path: str, list(str)
                Path to the RavenPro selection table(s). Use of wildcards is allowed.
            tax: korus.tax.AcousticTaxonomy
                Annotation taxonomy
            sep: str
                Character used to separate columns in the selection table. If None, tries to auto-detect the separator character.
            granularity: str
                Default granularity of annotations. Optional.
            timestamp_parser: callable
                Given a filename (str) returns the UTC timestamp embedded 
                in the filename as a datetime object. Optional.
            interactive: bool
                If True (default), the user will be prompted via the console to fix any invalid data.
                If False, entries with invalid data will be flagged by setting `valid=0`.
            progress_bar: bool
                Display progress bar. Default is False.

        Returns:
            df_out: pandas.DataFrame
                Table of selections

        Raises:
            FileNotFoundError: if the input file does not exist.
            ValueError: if the input table contains fields with invalid data types
    """
    # find files
    if isinstance(input_path, str):
        input_paths = glob(input_path)
    else:
        input_paths = [x for x in input_path if os.path.exists(x)]

    cprint(f"\n ## Found {len(input_paths)} selection tables matching the search criteria", "yellow")

    # load files
    dfs = []
    for input_path in input_paths:
        # detect separator
        if sep is None:
            with open(input_path, "r") as f:
                header = f.readline()
                if "\t" in header: sep = "\t"
                elif ";" in header: sep = ";"
                elif "," in header: sep = ","

        # load selections
        dfs.append(pd.read_csv(input_path, sep=sep))

    # concatenate
    df_in = pd.concat(dfs, ignore_index=True)

    # required columns
    required_cols = [
        "Begin File",
    ]

    # optional columns w/ default values
    optional_cols = {
        "Channel": 1,
        "File Offset (s)": 0,
        "Delta Time (s)": None,
        "Low Freq (Hz)": 0,
        "High Freq (Hz)": None,
        "Sound Source": None,
        "Sound Type": None,
        "Tentative Sound Source": None,
        "Tentative Sound Type": None,
        "Ambiguous Sound Source": None,
        "Ambiguous Sound Type": None,
        "Comments": None,
        "Valid": 1,
        "Tag": None,
    }

    has_duration = ("Delta Time (s)" in df_in.columns)
    has_freq_max = ("High Freq (Hz)" in df_in.columns)

    # check that required columns are present
    for c in required_cols:
        fail_msg = f"required column {c} missing in input file: {input_path}"
        assert c in df_in.columns, fail_msg

    # add default values for missing, optional columns
    for c,v in optional_cols.items():
        if c not in df_in.columns:
            df_in[c] = v

    # sort according to filename and start time
    df_in = df_in.sort_values(by=["Begin File", "File Offset (s)"])

    # define structure of output csv file
    data_out = {
        "path": [],
        "start_ms": [],
        "freq_min_hz": [],
        "sound_source": [],
        "sound_type": [],
        "tentative_sound_source": [],
        "tentative_sound_type": [],
        "ambiguous_sound_source": [],
        "ambiguous_sound_type": [],
        "granularity": [],
        "comments": [],
        "channel": [],
        "valid": [],
        "tag": [],
    }

    df_out = pd.DataFrame(data_out)

    # fill data into output dataframe
    df_out["path"] = df_in["Begin File"]
    df_out["start_ms"] = df_in["File Offset (s)"] * 1000
    df_out["freq_min_hz"] = df_in["Low Freq (Hz)"]
    df_out["sound_source"] = df_in["Sound Source"]
    df_out["sound_type"] = df_in["Sound Type"]
    df_out["tentative_sound_source"] = df_in["Tentative Sound Source"]
    df_out["tentative_sound_type"] = df_in["Tentative Sound Type"]
    df_out["ambiguous_sound_source"] = df_in["Ambiguous Sound Source"]
    df_out["ambiguous_sound_type"] = df_in["Ambiguous Sound Type"]
    df_out["comments"] = df_in["Comments"]
    df_out["channel"] = df_in["Channel"] - 1
    df_out["valid"] = df_in["Valid"]
    df_out["granularity"] = granularity
    df_out["tag"] = df_in["Tag"]
    
    # convert from comma-separated str to list object
    for col in ["tag", "ambiguous_sound_source", "ambiguous_sound_type"]:
        idx = ~df_out[col].isna()
        df_out.loc[idx, col] = df_out.loc[idx, col].apply(lambda x: x.split(","))
        df_out[col] = df_out[col].astype("object")
        df_out.loc[~idx, col] = None

    if has_duration:
        df_out["duration_ms"] = df_in["Delta Time (s)"] * 1000

    if has_freq_max:
        df_out["freq_max_hz"] = df_in["High Freq (Hz)"]

    # if the Selection Table contains the columns 'Granularity' and 'Batch Annotation', use them (in that order of preference)
    if "Granularity" in df_in.columns.values:
        df_out["granularity"] = df_in["Granularity"].apply(lambda x: x.lower() if isinstance(x, str) else granularity)

        if "Batch Annotation" in df_in.columns.values:
            warn_msg = "The selection table contains redundant columns `Granularity` and `Batch Annotation`. Dropping `Batch Annotation`."
            cprint(f"\n ## WARNING: {warn_msg}", "red")

    elif "Batch Annotation" in df_in.columns.values:
        idx_batch = (df_in["Batch Annotation"] == 1) 
        df_out.loc[idx_batch, "granularity"] = "batch"

    # replace NaN values
    df_out.comments = df_out.comments.fillna("")
    df_out.sound_source = df_out.sound_source.fillna("")
    df_out.sound_type = df_out.sound_type.fillna("")
    df_out.tentative_sound_source = df_out.tentative_sound_source.fillna("")
    df_out.tentative_sound_type = df_out.tentative_sound_type.fillna("")
    
    # validate data
    # also detect and parse ambiguous assignments in confident/tentative columns
    cprint(f"\n ## Validating labels ...", "yellow")

    df_out = validate_annotations(df_out, tax, interactive=interactive, progress_bar=progress_bar)

    # parse timestamps from filenames
    if timestamp_parser is not None:
        def get_start_utc(row):
            """ Helper function for obtaining UTC start times """
            return timestamp_parser(row.path) + timedelta(microseconds=row.start_ms * 1E3)

        df_out["start_utc"] = df_out.apply(lambda r: get_start_utc(r), axis=1)

    # dictionary to specify column types
    dtype_dict = {
        "path": "string",
        "start_ms": np.float32,
        "freq_min_hz": np.float32,
        "sound_source": "string",
        "sound_type": "string",
        "tentative_sound_source": "string",
        "tentative_sound_type": "string",
        "ambiguous_sound_source": "object",
        "ambiguous_sound_type": "object",
        "granularity": "string",
        "comments": "string",
        "channel": np.uint8,
        "tag": "object",
        "valid": np.uint8,
    }

    if has_duration:
        dtype_dict["duration_ms"] = np.float32

    if has_freq_max:
        dtype_dict["freq_max_hz"] = np.float32

    # cast columns, one at the time, for improved error handling
    for col_name, col_dtype in dtype_dict.items():
        try:
            if col_name in df_out.columns.values:
                df_out = df_out.astype({col_name: col_dtype})

        except:
            # if columns could not be casted to the desired type,
            # list the unique values in the column and their counts,
            # and issue a ValueError
            counts = df_out[col_name].value_counts(dropna=False)
            msg = (
                f"unable to cast {col_name} to {col_dtype} becase the "
                + "column contains entries with an invalid type; "
                + "column values are (value:count):"
            )
            for value, count in counts.items():
                msg += f" {value}:{count},"
            raise ValueError(msg)

    # replace empty sound source with root tag
    df_out.sound_source = df_out.sound_source.apply(lambda x: x if x != "" else tax.get_node(tax.root).tag)

    # replace empty tentative sound source with None
    df_out.tentative_sound_source = df_out.tentative_sound_source.apply(lambda x: x if x != "" else None)

    # replace empty sound type with root tag
    def fcn(r):
        if r.sound_type != "": 
            return r.sound_type
        else:
            try:
                types = tax.sound_types(r.sound_source)   
                return types.get_node(types.root).tag
            except AttributeError:
                return "Unknown"

    df_out.sound_type = df_out.apply(lambda r: fcn(r), axis=1)

    # replace empty tentative sound type with None
    df_out.tentative_sound_type = df_out.tentative_sound_type.apply(lambda x: x if x != "" else None)

    # round to appropriate number of digits
    decimals = {"start_ms": 0, "freq_min_hz": 1}
    if has_duration: 
        decimals["duration_ms"] = 0
    if has_freq_max:
        decimals["freq_max_hz"] = 1

    df_out = df_out.round(decimals)

    return df_out


def print_annotation_summary(conn, indices=None):
    """ Print annotation summary
    
        Args:
            conn: sqlite3.Connection
                Database connection
            indices: list(int)
                Indices in the annotation table. Optional.  
    """
    df = kdb.get_annotations(conn, indices)

    # validity
    counts = df.value_counts("valid", dropna=False).sort_index()
    counts = pd.DataFrame({"Count": counts})
    print(tabulate(counts, headers=["Valid", "Count"], tablefmt='psql'))

    # granularity
    counts = df.value_counts("granularity", dropna=False).sort_index()
    counts = pd.DataFrame({"Count": counts})
    print(tabulate(counts, headers=["Granularity", "Count"], tablefmt='psql'))

    # tags
    tag_count = unique_from_list(df.tag)
    counts = [[k,v] for k,v in tag_count.items()]
    print(tabulate(counts, headers=["Tag", "Count"], tablefmt='psql'))

    # confident labels
    counts = df.value_counts(["sound_source", "sound_type"], dropna=True)
    counts = pd.DataFrame({"Count": counts})
    print(tabulate(counts, headers=['Label', "Count"], tablefmt='psql'))

    # tentative labels
    counts = df.value_counts(["tentative_sound_source", "tentative_sound_type"], dropna=True)
    counts = pd.DataFrame({"Count": counts})
    print(tabulate(counts, headers=["Tentative label", "Count"], tablefmt='psql'))

    # ambiguous labels
    ambiguous_count = unique_from_list(df.ambiguous_label)
    counts = [[k,v] for k,v in ambiguous_count.items()]
    print(tabulate(counts, headers=["Ambiguous label", "Count"], tablefmt='psql'))


def validate_annotations(df, tax, interactive=True, progress_bar=False):
    """ Helper function for validating annotation data.
    
        Also parses ambiguous label assignments separated by slash (/).
        The function checks the confident/tentative label columns for occurrences of slash (/) characters.
        When ambiguous labels are found, they are moved to the ambiguous columns. 
        If they occur in the confident column, they are replaced by the last common ancestor node in the taxonomy. 
        If they occur in the tentative column, they are replaced by a 'null' value.

        Notes: Expects columns `ambiguous_sound_source`, `ambiguous_sound_type`, `tag` to be of type `object`.

        Args:
            df: pandas DataFrame
                Selection table
            tax: korus.tax.AcousticTaxonomy
                Annotation taxonomy
            interactive: bool
                If True (default), the user will be prompted via the console to fix any invalid data.
                If False, entries with invalid data will be flagged by setting `valid=0`.
            progress_bar: bool
                Display progress bar. Default is False.

        Returns:
            df: pandas DataFrame
                Selection table with ambiguous assignments resolved.
    """
    # maps for invalid labels
    src_map = dict()
    typ_map = dict()

    def _parse_ambiguous(labels, tax, label_map, idx, row, note=""):
        """ Helper function for parsing and validating ambiguous label assignments """
        if labels is None:
            return None, label_map
        
        labels_val = []
        for label in labels:
            if interactive:
                label, label_map = validate_label_interactive(label, tax, label_map, idx, row, note)
            else:
                label = validate_label(label, tax)

            if label is not None and label != "":
                labels_val.append(label)

        return labels_val, label_map

    # loop over all entries
    for idx,row in tqdm(df.iterrows(), disable=not progress_bar, total=df.shape[0]):   

        # repeat until all validations pass
        while True:

            try:
                # 1) parse ambiguous assignments in confident/tentative columns

                # sound source
                if row["ambiguous_sound_source"] is None:
                    s = row["sound_source"]
                    if "/" in s:
                        sources, src_map = _parse_ambiguous(s.split("/"), tax, src_map, idx, row)
                        parent = tax.last_common_ancestor(sources)
                        row["ambiguous_sound_source"] = sources
                        row["sound_source"] = parent

                    # tentative sound source
                    s = row["tentative_sound_source"]
                    if "/" in s:
                        sources, src_map = _parse_ambiguous(s.split("/"), tax, src_map, idx, row)
                        row["ambiguous_sound_source"] = sources
                        row["tentative_sound_source"] = ""

                if row["ambiguous_sound_type"] is None:
                    s = row["sound_source"]
                    typ_tax = tax.sound_types(s)

                    # sound type
                    t = row["sound_type"]
                    if "/" in t:            
                        types, typ_map = _parse_ambiguous(t.split("/"), typ_tax, typ_map, idx, row)                
                        parent = typ_tax.last_common_ancestor(types)
                        row["ambiguous_sound_type"] = types
                        row["sound_type"] = parent

                    # tentative sound type
                    t = row["tentative_sound_type"]
                    if "/" in t:
                        types, typ_map = _parse_ambiguous(t.split("/"), typ_tax, typ_map, idx, row)      
                        row["ambiguous_sound_type"] = types
                        row["tentative_sound_type"] = ""


                # 2) validate labels

                # 2a) sound source
                ss = row["sound_source"]
                if interactive:
                    row["sound_source"], src_map = validate_label_interactive(ss, tax, src_map, idx, row)
                else:
                    row["sound_source"] = validate_label(ss, tax)

                # 2b) tentative sound source
                tss = row["tentative_sound_source"]
                if interactive:
                    row["tentative_sound_source"], src_map = validate_label_interactive(tss, tax, src_map, idx, row)
                else:
                    row["tentative_sound_source"] = validate_label(tss, tax)

                # 2c) ambiguous sound source
                ass = row["ambiguous_sound_source"]
                ambiguous_sources, src_map = _parse_ambiguous(ass, tax, src_map, idx, row)
                row["ambiguous_sound_source"] = ambiguous_sources

                # 2d) sound type
                st = row["sound_type"]

                # pick which sound-type tree to validate against
                typ_tree = tax.sound_types(ss)
                note = f" (for sound source `{ss}`)"

                if interactive:
                    row["sound_type"], typ_map = validate_label_interactive(st, typ_tree, typ_map, idx, row, note)
                else:
                    row["sound_type"] = validate_label(st, typ_tree)

                # 2e) tentative sound type

                tst = row["tentative_sound_type"]

                # pick which sound-type tree to validate against
                if tss is not None and tss != "":
                    typ_tree = tax.sound_types(tss)
                    note = f" (for sound source `{tss}`)"
                else:
                    typ_tree = tax.sound_types(ss)
                    note = f" (for sound source `{ss}`)"

                if interactive:                    
                    row["tentative_sound_type"], typ_map = validate_label_interactive(tst, typ_tree, typ_map, idx, row, note)
                else:
                    row["tentative_sound_type"] = validate_label(tst, typ_tree)

                # 2f) ambiguous sound type
                ast = row["ambiguous_sound_type"]

                # pick which sound-type tree to validate against
                if ambiguous_sources is None:
                    typ_tree = tax.sound_types(ss)
                    note = f" (for sound source `{ss}`)"
                else:
                    parent = tax.last_common_ancestor(ambiguous_sources)
                    typ_tree = tax.sound_types(parent)
                    note = f" (for sound source `{parent}`)"

                types, typ_map = _parse_ambiguous(ast, typ_tree, typ_map, idx, row, note=note)
                row["ambiguous_sound_type"] = types


                # 3) all validations have passed so exit the loop
                df.loc[idx] = row                       
                break        


            except ValueError as e:
                if interactive:
                    # edit row manually
                    row = edit_row_manually(idx, row)

                else:
                    # flag as containing invalid data
                    row.valid = 0
                    df.loc[idx] = row

                    warn_msg = f"{str(e)}. Entry {idx} flagged as containing invalid data."
                    cprint(f"\n ## WARNING: {warn_msg}", "red")

                    break

    # replace NaN values
    df.comments = df.comments.fillna("")
    df.sound_source = df.sound_source.fillna("")
    df.sound_type = df.sound_type.fillna("")
    df.tentative_sound_source = df.tentative_sound_source.fillna("")
    df.tentative_sound_type = df.tentative_sound_type.fillna("")

    return df


def edit_row_manually(idx, row):
    """ Edit an annotation manually.

        The annotation data is saved to a temporary YAML file. The user is prompted via the console 
        to edit and save the file, then hit ENTER to proceed.
        
        Datetime objects are converted to strings using the format `%Y-%m-%d %H:%M:%S.%f`

        Args:
            idx: int
                Index
            row: pandas Series
                Values

        Returns:
            row: pandas Series
                The edited values. Returns None if the user elects to ignore/skip the entry.
    """
    path = f"korus-entry-{idx}.yaml"

    with open(path, "w") as f:
        row_dict = row.to_dict()
        for k,v in row_dict.items():
            if isinstance(v, datetime):
                row_dict[k] = v.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]

        yaml.dump(row_dict, f)      

    try:
        msg = f"\n >> Entry {idx} saved to {path}. Edit file manually and save. Then hit ENTER to proceed with submission. Or type i/ignore to skip the entry."
        cprint(msg, "yellow")
        res = input()
        if res.lower() in ["i", "ignore"]:
            return None
    
    except KeyboardInterrupt:
        os.remove(path)
        raise

    with open(path, "r") as f:
        row_dict = yaml.safe_load(f)

    row = pd.Series(row_dict)

    os.remove(path)

    return row


def validate_label(x, tax):
    """ Validate a label against a taxonomy of allowed values.

        Args:
            x: str
                Label to be validated. OBS: If None or empty string, no validation is performed.
            tax: Korus.tax.AcousticTaxonomy
                Taxonomy of allowed labels, against which x will be validated

        Returns:
            x: str
                The validated label.

        Raises:
            ValueError: if the label is found to be invalid
    """
    if x is None or x == "":
        return x
    
    if tax.get_id(x) is None:
        raise ValueError(f"Invalid label `{x}`")
    
    return x


def validate_label_interactive(x, tax, label_map, idx, row, note=""):
    """ Interactive session for validating a label against a taxonomy of allowed values.

        If the label is found to be invalid, the user is prompted via the terminal for an alternative label.

        Args:
            x: str
                Label to be validated
            tax: Korus.tax.AcousticTaxonomy
                Taxonomy of allowed labels, against which x will be validated
            label_map: dict
                Mapping for invalid labels
            idx: int
                Row index. Only used for interactive prompt.
            row: pandas Series
                Row values. Only used for interactive prompt.
            note: str
                Optional note, appended to the console message.

        Returns:
            y: str
                The validated label.
            label_map: dict
                The updated label map.

        Raises:
            ValueError: if the user requests to switch to manual editing mode
    """
    if x is None or x == "":
        return x, label_map

    y = x
    while tax.get_id(y) is None:
        if y in label_map:
            y = label_map[y]

        else:
            try:
                cprint(f" >> Invalid label in entry {idx}: {y} {note}", "red")
                msg = " >> Options:"
                msg += "\n >>  - [name]:      alternative, valid label"
                msg += "\n >>  - v/view:      view entry"
                msg += "\n >>  - t/taxonomy:  view taxonomy"
                msg += "\n >>  - m/manual:    switch to manual editing mode"
                msg += "\n >>  - Ctrl-C:      abort\n"

                res = input(msg)

                if res in ["v","view"]:
                    tbl = [[k,v] for k,v in row.to_dict().items()]
                    print(tabulate(tbl, headers=["Field","Value"], tablefmt="psql"))                    

                elif res in ["t","taxonomy"]:
                    tax.show()

                elif res in ["m","manual"]:
                    raise ValueError

                else:
                    y = res

            except KeyboardInterrupt:
                raise

    if y != x:
        cprint(f" >> Replacing label in entry {idx}: {x} -> {y}", "green")
        
        if x not in label_map:
            replace = ui.UserInputYesNo(
                "label_replacement_rule", 
                "Apply same replacement rule to all other entries? [y/N]", 
            ).request()
            if replace:
                label_map[x] = y

    return y, label_map





import os
import pytest
import korus.tax as kx
from korus.tree import Tree

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(current_dir, "assets")
path_to_tmp = os.path.join(path_to_assets, "tmp")


def test_taxonomy():
    """ Test basic functionalities of the Taxonomy class. Not a comprehensive test."""
    sqlite_path = os.path.join(path_to_tmp, "mytax.sqlite")
    if os.path.exists(sqlite_path):
        os.remove(sqlite_path)

    # check that we can initialize the class
    tax = kx.Taxonomy(name="mytax", root_tag="base", path=sqlite_path)

    # check that we can create some nodes
    tax.create_node("PC", parent="base", name="Pulsed Call")
    tax.create_node("BZ", parent="base")
    tax.create_node("S1", parent="PC")
    tax.create_node("S2", parent="PC", name="an S2 call")
    tax.create_node("S3", parent="PC")
    tax.create_node("S4", parent="PC")
    tax.create_node("S5", parent="PC")

    n = tax.get_node("S2")
    assert n.tag == "S2"
    assert n.data["name"] == "an S2 call"

    # if we attempt to merge two 'new' nodes, Korus complains
    with pytest.raises(RuntimeError):
        tax.merge_nodes("S23", children=["S2","S3"])

    # check that we are keeping track of created and deleted nodes
    assert len(tax.removed_nodes) == 0
    assert len(tax.created_nodes) == 8 #7 + 1 root node

    count = 0
    for k,v in tax.created_nodes.items():
        # check that root node has created_from = None and is_equivalent = False
        if count == 0:
            assert v[0] == [None]
            assert not v[1]

        # check that the next two nodes have created_from = root note and is_equivalent = False
        elif count == 1 or count == 2:
            assert v[0] == [tax.get_id("base")]
            assert v[1] == False

        # check that the remaining nodes have created_from = PC and is_equivalent = False
        elif count > 2:
            assert v[0] == [tax.get_id("PC")]
            assert v[1] == False

        count += 1

    # check that we are able to save the taxonomy and that it is assigned version no. 1
    tax.save()
    assert tax.latest_version() == 1

    # check we can now merge nodes
    tax.merge_nodes("S23", children=["S2","S3"])
    tax.merge_nodes("S45", children=["S4","S5"], remove=True)

    # check that we are keeping correct track of created/removed nodes
    assert len(tax.created_nodes) == 2
    assert len(tax.removed_nodes) == 2

    for k,v in tax.created_nodes.items():
        n = tax.get_node(k)

        if n.tag == "S23":
            assert len(v[0]) == 2
            assert v[0] == [tax.get_id("S2"), tax.get_id("S3")]
            assert v[1] == True

        elif n.tag == "S45":
            assert len(v[0]) == 2
            assert v[1] == True

    for k,v in tax.removed_nodes.items():
        assert v[0] == [tax.get_id("S45")]
        assert v[1] == False

    assert tax.version == 1

    tax.save()
    assert tax.latest_version() == 2

    #check that we can load the taxonomy back into memory
    tax = kx.Taxonomy.load(sqlite_path, name="mytax", version=2)
    assert tax.version == 2
    assert tax.path == sqlite_path


def test_acoustic_taxonomy():
    """ Test basic functionalities of the AcousticTaxonomy class. Not a comprehensive test."""
    sqlite_path = os.path.join(path_to_tmp, "myacoutax.sqlite")
    if os.path.exists(sqlite_path):
        os.remove(sqlite_path)

    # check that we can initialize the class
    tax = kx.AcousticTaxonomy(name="myacoutax", path=sqlite_path)

    # check that we can create some sound sources
    tax.create_sound_source("Bio", name="Biological sound source")
    tax.create_sound_source("Mammal", parent="Bio", name="Any mammal")
    kw = tax.create_sound_source("KW", parent="Mammal")
    kw = tax.create_sound_source("Dolphin", parent="Mammal")
    kw = tax.create_sound_source("HW", parent="Mammal")

    n = tax.get_node("HW")
    assert n.tag == "HW"
    assert tax.parent(tax.get_id("HW")).tag == "Mammal"

    # and some sound types
    tax.create_sound_type("TC", "Mammal", name="Tonal Call")
    tax.create_sound_type("CK", "Dolphin", name="Click")
    tax.create_sound_type("CK", "KW", name="Click")
    tax.create_sound_type("PC", "KW", name="Pulsed Call")

    t = tax.sound_types("KW")
    n = t.get_node("PC")
    assert n.tag == "PC"
    assert n.data["name"] == "Pulsed Call"
    assert t.parent(t.get_id("PC")).tag == "Unknown"

    # add another source, after adding sound types
    srkw = tax.create_sound_source("SRKW", parent="KW")

    # check that it inherited the sound types
    t_kw = tax.sound_types("KW")
    t_srkw = tax.sound_types("SRKW")
    for nid in t_kw.expand_tree(mode=Tree.DEPTH):
        n_srkw = t_srkw.get_node(nid)
        assert n_srkw.identifier == nid

    # add more sound types
    tax.create_sound_type("S01", "SRKW", parent="PC", name="S1 call")
    tax.create_sound_type("S02", "SRKW", parent="PC", name="S2 call")
    tax.create_sound_type("S03", "SRKW", parent="PC", name="S3 call")
    tax.create_sound_type("TCX", "SRKW", parent="TC", name="Tonal call X")

    # check that we can save the taxonomy
    tax.save()
    assert tax.latest_version() == 1

    # check that we can merge sound sources
    tax.merge_sound_sources(
        "Toothed", 
        children=["KW","Dolphin"], 
        remove=False, 
        data_merge_fcn=None,
        name="Odontecetes"
    )

    # check that toothed whales inherit sound types from parent node (Mammal)
    t_mammal = tax.sound_types("Mammal")
    t_toothed = tax.sound_types("Toothed")
    for nid in t_mammal.expand_tree(mode=Tree.DEPTH):
        n_toothed = t_toothed.get_node(nid)
        assert n_toothed.identifier == nid

    # check that we can merge sound types
    tax.merge_sound_types(
        tag="S2&3",
        source_tag="SRKW", 
        children=["S02","S03"], 
        remove=False, 
        data_merge_fcn=None,
    )

    # save version 2
    tax.save()
    assert tax.latest_version() == 2

    # check that we can load the taxonomy back into memory
    tax = kx.AcousticTaxonomy.load(sqlite_path, name="myacoutax", version=1)
    assert tax.version == 1

    # verify that sound types are as expected
    sound_types = tax.sound_types("SRKW")
    expected = ["Unknown","CK","PC","S01","S02","S03","TC","TCX"]
    for i,nid in enumerate(sound_types.expand_tree(mode=Tree.DEPTH)):
        assert sound_types.get_node(nid).tag == expected[i]

    # check that we can ascend up through the sound-type tree
    gen = tax.ascend("SRKW", "S01", include_start_node=False)
    expected = [
        ("SRKW","PC"),
        ("SRKW","Unknown"),
        ("KW","PC"),
        ("KW","Unknown"),
        ("Mammal","Unknown"),
        ("Bio","Unknown"),
        ("Unknown","Unknown"),      
    ]
    for i,(s,t) in enumerate(gen):
        assert (s, t) == expected[i]

    gen = tax.ascend("SRKW", include_start_node=False)
    expected = ["KW","Mammal","Bio","Unknown"]
    for i,(s,t) in enumerate(gen):
        assert s == expected[i]

    # check that we can descend down through the sound-type tree
    gen = tax.descend("Mammal", "TC", include_start_node=True)
    expected = [
        ("Mammal","TC"),
        ("Dolphin","TC"),
        ("HW","TC"),
        ("KW","TC"),
        ("SRKW","TC"),
        ("SRKW","TCX"),      
    ]
    for i,(s,t) in enumerate(gen):
        assert (s, t) == expected[i]

    gen = tax.descend("KW", include_start_node=False)
    expected = ["SRKW"]
    for i,(s,t) in enumerate(gen):
        assert s == expected[i]


    tax.create_sound_type("S99", "SRKW", parent="PC", name="S99 call")
    tax.save("added S99 call")

    tax.create_sound_type("S100", "SRKW", parent="PC", name="S100 call")
    tax.save("added S100 call", overwrite=True)

    tax.create_sound_type("S101", "SRKW", parent="PC", name="S101 call")
    tax.save("added S101 call")

    tax.create_sound_type("S102", "SRKW", parent="PC", name="S102 call")
    tax.save("added S102 call")

    tax = kx.AcousticTaxonomy.load(sqlite_path, name="myacoutax", version=4)
    tax.create_sound_type("S103", "SRKW", parent="PC", name="S103 call")
    tax.save("added S103 call", overwrite=True)

    import sqlite3
    conn = sqlite3.connect(sqlite_path)
    c = conn.cursor()
    rows = c.execute("SELECT id,name,version,comment FROM taxonomy").fetchall()
    rows[0] == (1, 'myacoutax', 1, None)
    rows[1] == (2, 'myacoutax', 2, None)
    rows[2] == (3, 'myacoutax', 3, 'added S99 call; added S100 call')
    rows[3] == (4, 'myacoutax', 4, 'added S101 call; added S103 call')
    rows[4] == (5, 'myacoutax', 5, 'added S102 call')


    rows = c.execute(f"SELECT * FROM taxonomy_created_node WHERE taxonomy_id = {4}").fetchall()
    print(rows)

    conn.close()


import os
import json
import warnings
import logging
import sqlite3
from datetime import datetime
from treelib import Tree
import korus.tree as ktr
import korus.db as kdb 
import korus.db_util.table as ktb


class Taxonomy(ktr.KTree):
    """ Class for managing annotation taxonomies with a tree-like 
        structure where every child nodes has precisely one parent 
        node.

        The Taxonomy class is derived from korus.tax.KTree, 
        adding the following functionalities,

         * SQLite database storage
         * version tracking, including node ancestry       

        TODO: add assertion to check if there is an open connection to the sqlite database
        
        Args:
            name: str
                Short, descriptive name for the taxonomy.
            root_tag: str
                Tag for the root node. If specified, the root node will be 
                automatically created at initialisation.
            path: str
                Path to a SQLite database file for storing the taxonomy. 
                If None, the path will be set to ./{@name}.sqlite.
            overwrite: bool
                Whether to overwrite the file if it already exists.
    """
    def __init__(self, name="taxonomy", root_tag="root", path=None, overwrite=False):
        self.name = name
        self.version = None
        self.path = f"./{name}.sqlite" if path is None else path
        if os.path.exists(self.path) and overwrite:
            os.remove(self.path)

        super().__init__(root_tag=root_tag)

    @classmethod
    def from_dict(cls, input_dict, data_transform=None, path=None):
        """ Load a taxonomy from a dictionary.

            Expects the dictionary to have the keys 'name', 'version', 'tree'.

            Within the dictionary, the key 'children' is used to designate branching 
            points, and the key 'data' is used to designate any data associated with a 
            node.

            Args:
                input_dict: dict()
                    Input dictionary.
                data_transform: callable
                    This function gets applied to all entries in the dictionary with the key 'data'.
                path: str
                    Path to a SQLite database file for storing the taxonomy. 
            
            Returns:
                tax: korus.tax.Taxonomy
                    The taxonomy
        """
        tree_dict = input_dict.pop("tree")
        version = input_dict.pop("version")
        tax = ktr.tree_from_dict(cls(**input_dict, root_tag=None, path=path), recipe=tree_dict, data_transform=data_transform)
        tax.version = version
        return tax

    @classmethod
    def load(cls, path, name="taxonomy", version=None):
        """ Load an existing taxonomy from an SQLite database.

            The method expects to find the table,
            
                taxonomy(
                    id INTEGER NOT NULL,
                    name TEXT NOT NULL,
                    version TEXT,
                    tree JSON NOT NULL,
                    timestamp TEXT,
                    comment TEXT,
                    PRIMARY KEY (id),
                    UNIQUE (name, version)
                )
        
            Args:
                path: str
                    Path to the database file (.sqlite)
                name: str
                    Taxonomy name.
                version: int
                    Taxonomy version. If not specified, the latest version will be loaded.
                
            Returns:
                : korus.tax.Taxonomy
                    The taxonomy
        """
        conn = sqlite3.connect(path)
        
        query = f"SELECT name,version,tree FROM taxonomy WHERE name = '{name}'"
        if version is not None:
            query += f" AND version = '{version}'"
        
        c = conn.cursor()
        rows = c.execute(query).fetchall()

        conn.close()

        rows = sorted(rows, key=lambda x: x[1]) #sort according to version no.

        (name,version,data) = rows[-1]

        tax_dict = {"name":name, "version":version, "tree":json.loads(data)}
        return cls.from_dict(tax_dict, path=path)
    
    def latest_version(self):
        """ Get the latest version number"""
        if os.path.exists(self.path):
            conn = sqlite3.connect(self.path)
            c = conn.cursor()
            query = f"SELECT version FROM taxonomy WHERE name = '{self.name}'"
            rows = c.execute(query).fetchall()
            conn.close()
            version_numbers = [int(row[0]) for row in rows]
            if len(version_numbers) == 0:
                return None
            else:
                return max(version_numbers)

        else:
            return None

    def to_dict(self):
        """ Transform the taxonomy to a dictionary.
    
            Returns:
                : dict()
                    The taxonomy in the form of a dictionary.
        """
        return {"name":self.name, "version":self.version, "tree":ktr.tree_to_dict(self)}

    def save(self, comment=None, overwrite=False):
        """ Save the taxonomy.

            The version number is automatically incremented by +1 when this method is 
            called, unless @overwrite is set to True in which case the currently loaded 
            version is overwritten.

            Args:
                comment: str
                    Optional field. Typically used for describing the main changes made 
                    to the taxonomy since the last version.
                overwrite: bool
                    If True, the version no. is not incremented and instead the currently 
                    loaded version is overwritten.
        """
        v = self.latest_version()

        if v is None:
            self.version = 1 
        else:
            if not overwrite:
                self.version = v + 1

        self._to_sqlite(self.path, comment, overwrite)
        self.clear_history()

    def _to_sqlite(self, path, comment, overwrite=False):
        """ Save the taxonomy to an SQLite database.

            TODO: Implement @overwrite

            The database must contain the tables,

                taxonomy(
                    id INTEGER NOT NULL,
                    name TEXT NOT NULL,
                    version TEXT NOT NULL,
                    tree JSON NOT NULL,
                    timestamp TEXT,
                    comment TEXT,
                    PRIMARY KEY (id),
                    UNIQUE (name, version)
                )

                taxonomy_created_node(
                    id TEXT NOT NULL,
                    precursor_id JSON NOT NULL,
                    is_equivalent INTEGER NOT NULL,
                    taxonomy_id INTEGER NOT NULL,
                    PRIMARY KEY (id),
                    FOREIGN KEY (taxonomy_id) REFERENCES taxonomy (id)
                )

                taxonomy_removed_node(
                    id TEXT NOT NULL,
                    inheritor_id JSON NOT NULL,
                    is_equivalent INTEGER NOT NULL,
                    taxonomy_id INTEGER NOT NULL,
                    PRIMARY KEY (id),
                    FOREIGN KEY (taxonomy_id) REFERENCES taxonomy (id)
                )
                
            If the database does not already exist or it does not 
            contain the required table, they will be automatically 
            created.

            Args:
                path: str
                    Path to the SQLite database file (*.sqlite)
                comment: str
                    Optional field. Typically used for describing the main changes made 
                    to the taxonomy since the last version.
                overwrite: bool
                    Set to True to allow existing entries in the taxonomy table with the 
                    same name and version no. to be overwritten.
        """
        db_exists = os.path.exists(path)

        # if folder does not exist, create it
        dir_path = os.path.dirname(path)
        if not os.path.exists(dir_path):
            os.makedirs(dir_path)

        conn = sqlite3.connect(path) #will create the *.sqlite file if it does not already exist
        c = conn.cursor()

        # create the tables if they do not already exist
        if not db_exists:
            ktb.create_taxonomy_table(conn)
            ktb.create_taxonomy_created_node_table(conn)
            ktb.create_taxonomy_removed_node_table(conn)

        tax_dict = self.to_dict()

        timestamp = datetime.utcnow().strftime("%Y-%m-%d %H:%M:%S")

        # check if taxonomy already exists
        name = tax_dict["name"]
        version = tax_dict["version"]
        rows = c.execute(f"SELECT id FROM taxonomy WHERE name = '{name}' AND version = {version}").fetchall()
        if len(rows) == 1:
            tax_exists = True
            tax_id = rows[0][0]
        else:
            tax_exists = False
            tax_id = None
            prev_comment = None

        if tax_exists and not overwrite:
            err_msg = f"Failed to save taxonomy."\
                + f" Database already contains a taxonomy named {name} with version no. {version}."\
                + f" Set @overwrite to True if you wish to overwrite the existing entry."
            logging.error(err_msg)
            return c

        elif tax_exists and overwrite:
            # retrieve comment
            (prev_comment,) = c.execute(f"SELECT comment FROM taxonomy WHERE id = {tax_id}").fetchall()[0]
            # delete existing entry
            c.execute(f"DELETE FROM taxonomy WHERE id = {tax_id}")

        # append to previous comment, if overwriting
        if prev_comment is not None:
            comment = prev_comment + "; " + comment

        # save taxonomy
        c.execute(
            "INSERT INTO taxonomy VALUES (?, ?, ?, ?, ?, ?)",
            [tax_id, tax_dict["name"], tax_dict["version"], json.dumps(tax_dict["tree"]), timestamp, comment]
        ) 

        if tax_id is None:
            tax_id = c.lastrowid 

        # save precursors of created nodes
        for created_id, (precursor_id, is_equivalent) in self.created_nodes.items():
            c.execute("INSERT INTO taxonomy_created_node VALUES (?, ?, ?, ?)", [created_id, json.dumps(precursor_id), int(is_equivalent), tax_id])

        # save inheritors of removed nodes
        for removed_id, (inheritor_id, is_equivalent) in self.removed_nodes.items():
            c.execute("INSERT INTO taxonomy_removed_node VALUES (?, ?, ?, ?)", [removed_id, json.dumps(inheritor_id), int(is_equivalent), tax_id])

        conn.commit()
        conn.close()


class AcousticTaxonomy(Taxonomy):
    """ Class for managing annotation acoustic taxonomies with a nested, 
        tree-like structure.
        
        When annotating acoustic data, it is customary to describe sounds
        by their source (e.g. a killer whale) as well as their type, i.e, 
        their aural and spectral characteristics (e.g. a tonal call). 
        
        In the AcousticTaxonomy class, the nodes of the (primary) tree are 
        the sound sources, and nested within each of these node is a 
        (secondary) tree of sound types.

        Args:
            name: str
                Descriptive, short name for the taxonomy.
            root_tag: str
                Tag for the root node. If specified, the root node will be 
                automatically created at initialisation.
            version: int
                Version number.
            overwrite: bool
                Whether to overwrite the file if it already exists.
    """
    def __init__(
            self, 
            name="acoustic_taxonomy", 
            root_tag="Unknown", 
            path=None,
            overwrite=False,
        ):
        self._type_root_tag = root_tag
        super().__init__(name=name, root_tag=root_tag, path=path, overwrite=overwrite)

    @classmethod
    def from_dict(cls, input_dict, path=None):
        """ Load an acoustic taxonomy from a dictionary.

            Overwrites Taxonomy.from_dict

            Expects the dictionary to have the keys 'name', 'version', 'tree'.

            Within the dictionary, the key 'children' is used to designate branching 
            points, and the key 'data' is used to designate any data associated with a 
            node. The key 'types' is used to designate a sub-tree of sound types 
            associated with a particular node. 

            Args:
                input_dict: dict()
                    Input dictionary.
                path: str
                    Path to a SQLite database file for storing the taxonomy. 
        """
        def data_transform(x):
            if x is not None:
                types_dict = x.pop("sound_types", None)
                if types_dict is not None:
                    tree = ktr.KTree(root_tag=None)
                    tree = ktr.tree_from_dict(tree, types_dict)
                    x.update({"sound_types": tree})

            return x

        return super().from_dict(input_dict, data_transform, path=path)

    def _to_sqlite(self, path, comment=None, overwrite=False):
        """ Save the taxonomy to an SQLite database.

            Overwrites Taxonomy._to_sqlite

            The database must contain the tables,

                taxonomy(
                    id INTEGER NOT NULL,
                    name TEXT NOT NULL,
                    version TEXT,
                    tree JSON NOT NULL,
                    timestamp TEXT,
                    comment TEXT,
                    PRIMARY KEY (id),
                    UNIQUE (name, version)
                )

                taxonomy_created_node(
                    id TEXT NOT NULL,
                    precursor_id JSON NOT NULL,
                    is_equivalent INTEGER NOT NULL,
                    taxonomy_id INTEGER NOT NULL,
                    PRIMARY KEY (id),
                    FOREIGN KEY (taxonomy_id) REFERENCES taxonomy (id)
                )

                taxonomy_removed_node(
                    id TEXT NOT NULL,
                    inheritor_id JSON NOT NULL,
                    is_equivalent INTEGER NOT NULL,
                    taxonomy_id INTEGER NOT NULL,
                    PRIMARY KEY (id),
                    FOREIGN KEY (taxonomy_id) REFERENCES taxonomy (id)
                )
                
                label(
                    id INTEGER NOT NULL,
                    taxonomy_id INTEGER NOT NULL,
                    sound_source_tag TEXT,
                    sound_source_id TEXT,
                    sound_type_tag TEXT,
                    sound_type_id TEXT,
                    PRIMARY KEY (id),
                    FOREIGN KEY (taxonomy_id) REFERENCES taxonomy (id),
                    UNIQUE (taxonomy_id, sound_source_uuid, sound_type_uuid)
                )

            If the database does not already exist or it does not 
            contain the required tables, they will be automatically 
            created.

            Args:
                path: str
                    Path to the SQLite database file (*.sqlite)
                comment: str
                    Optional field. Typically used for describing the main changes made 
                    to the taxonomy since the last version.
                overwrite: bool
                    Set to True to allow existing entries in the taxonomy table with the 
                    same name and version no. to be overwritten.
        """
        db_exists = os.path.exists(path)

        conn = sqlite3.connect(path) #will create the *.sqlite file if it does not already exist

        # create the tables if they do not already exist
        if not db_exists:
            ktb.create_taxonomy_table(conn)
            ktb.create_taxonomy_created_node_table(conn)
            ktb.create_taxonomy_removed_node_table(conn)
            ktb.create_label_table(conn)

        # insert data into tables
        kdb.insert_taxonomy(conn, self, comment=comment, overwrite=overwrite)

        conn.commit()
        conn.close()

    def create_node(self, tag, identifier=None, parent=None, precursor=None, inherit_types=True, **kwargs):
        """ Add a sound source to the taxonomy.

            Overwrites KTree.create_node.

            Sound source attributes can be specified using keyword arguments.

            It is recommended to include the following attributes:

                * name
                * description
                * scientific_name
                * tsn

            Args:
                tag: str
                    Tag for the sound source.
                parent: str
                    Parent sound source identifier or tag. The default value is None, implying 'root' as parent.
                precursor: str, list(str)
                    Used for tracking the ancestry of the child node. If None, the parent identifier will be used.
                inherit_types: bool
                    Inherit sound types from parent source. Default is True.

            Returns: 
                node: treelib.node.Node
                    The new node object
                    
            Raises:
                AssertionError: if the taxonomy already contains a sound source with the specified tag
        """
        if parent is None:
            parent = self.root

        if kwargs is None or len(kwargs) == 0:
            kwargs = dict()

        if "sound_types" not in kwargs:
            if inherit_types and self.root is not None: # inherit sound-type tree from parent            
                kwargs["sound_types"] = self.get_node(parent).data["sound_types"].deepcopy()
            
            else: # create empty sound-type tree            
                kwargs["sound_types"] = ktr.KTree(root_tag=self._type_root_tag)

        return super().create_node(tag=tag, identifier=identifier, parent=parent, precursor=precursor, **kwargs)

    def create_sound_source(self, tag, parent=None, precursor=None, inherit_types=True, **kwargs):
        """ Merely a wrapper for :meth:`create_node`"""
        return self.create_node(tag, parent=parent, precursor=precursor, inherit_types=inherit_types, **kwargs)
    
    def create_sound_type(self, tag, source_tag=None, parent=None, recursive=True, **kwargs):
        """ Add a sound type to the taxonomy.

            Note that the sound type must be associated with a particular sound source, i.e., 
            a particular node in the primary tree (which can be the root node). 
            
            Also, note that if @recursive is set to true, all child nodes (sound sources) 
            will inherit the sound type.

            Keyword arguments can be used to specify additional data to be associated 
            with the sound type, e.g., a wordy description of its acoustic characteristics.

            Args:
                tag: str
                    Tag for the sound type. Must be unique within the sound source.
                source_tag: str
                    Tag of the sound source that the sound type is to be associated with.
                parent: str
                    Tag or identifier of the parent sound type. Use this to create a hierarchy 
                    of sound types.
                recursive: bool
                    Also add this sound type to all descendant sound sources. Default is True.
        """
        source_id = self.get_id(source_tag)

        if recursive:
            source_ids = self.expand_tree(nid=source_id, mode=Tree.DEPTH)
        else:
            source_ids = [source_id]

        # loop over sound sources
        for source_id in source_ids:
            n = self.get_node(source_id)
            types = n.data["sound_types"]

            if parent is None:
                parent_tag = types.root
            else:
                parent_tag = parent

            parent_id = types.get_id(parent_tag)

            if kwargs is None or len(kwargs) == 0:
                kwargs = dict()
            
            # add the new sound type to the KTree
            types.create_node(tag=tag, parent=parent_id, **kwargs)

    def merge_sound_sources(self, tag, children=None, remove=False, data_merge_fcn=None, inherit_types=True, **kwargs):
        """ Merge sound sources """
        return self.merge_nodes(tag, children=children, remove=remove, data_merge_fcn=data_merge_fcn, inherit_types=inherit_types, **kwargs)
    
    def merge_sound_types(self, tag, source_tag=None, children=None, remove=False, 
                          data_merge_fcn=None, recursive=True, **kwargs):
        """ Merge sound types """
        source_id = self.get_id(source_tag)

        if recursive:
            source_ids = self.expand_tree(nid=source_id, mode=Tree.DEPTH)
        else:
            source_ids = [source_id]

        for source_id in source_ids:
            types = self.get_node(source_id).data["sound_types"]
            types.merge_nodes(tag, children=children, remove=remove, data_merge_fcn=data_merge_fcn)

    def clear_history(self):
        """ Overwrites Taxonomy.clear_history
        """
        # commit changes for sound sources
        super().clear_history()

        # commit changes for sound types
        for source_id in self.expand_tree(mode=Tree.DEPTH):
            self.get_node(source_id).data["sound_types"].clear_history()

    def sound_types(self, source_tag):
        """ Returns the KTree of sound types associated with a given sound source
        
        Args:
            source_tag: str
                Sound source tag

        Returns:
            t: korus.tree.KTree
                Sound types. Returns None if the sound source does not exist.
        """
        try:
            return self.get_node(source_tag).data["sound_types"]
        except:
            return None

    @property
    def created_nodes(self):
        """Overwrites KTree.created_nodes"""
        created_nodes = super().created_nodes
        for source_id in self.expand_tree(mode=Tree.DEPTH):
            created_nodes.update(self.get_node(source_id).data["sound_types"].created_nodes)
        
        return created_nodes

    @property
    def removed_nodes(self):
        """Overwrites KTree.removed_nodes"""
        removed_nodes = super().removed_nodes
        for source_id in self.expand_tree(mode=Tree.DEPTH):
            removed_nodes.update(self.get_node(source_id).data["sound_types"].removed_nodes)
        
        return removed_nodes

    def ascend(self, source_tag, type_tag=None, include_start_node=True):
        """ Returns a python generator for ascending the taxonomy starting 
            at @source_tag, @type_tag.

            Args:
                source_tag: str
                    Sound source tag or identifier of starting node.        
                type_tag: str
                    Sound type tag of starting node. If None or '%', the 
                    generator will only iterate through the sound-source nodes.
                include_start_node: bool
                    Whether to include the starting node. Default is True.

            Yields:
                source_tag, type_tag: str, str
        """
        debug_msg = f"[{self.__class__.__name__}] Ascending {self.name} v{self.version} starting from ({source_tag},{type_tag})"
        logging.debug(debug_msg)

        types = self.get_node(source_tag).data["sound_types"]  #sound-type tree
        source_gen = self.rsearch(source_tag)  #ascending source-id generator

        counter = 0
        for sid in source_gen: #ascend up through sound sources
            source_i = self.get_node(sid)
            types_i = source_i.data["sound_types"]

            if type_tag is None or type_tag == "%":
                if include_start_node or counter > 0:
                    yield source_i.tag, type_tag
                
                counter += 1

            else:
                for tid in types.rsearch(type_tag):
                    if types_i.get_node(tid) is not None:
                        type_gen = types_i.rsearch(tid)
                        break

                for tid in type_gen:
                    type_i = types_i.get_node(tid)
                    if include_start_node or counter > 0:
                        yield source_i.tag, type_i.tag
                    
                    counter += 1

    def descend(self, source_tag, type_tag=None, include_start_node=True):
        """ Returns a python generator for descending the taxonomy starting 
            at @source_tag, @type_tag.

            Args:
                source_tag: str
                    Sound source tag or identifier of starting node.        
                type_tag: str
                    Sound type tag of starting node. If None or '%', the 
                    generator will only iterate through the sound-source nodes.
                include_start_node: bool
                    Whether to include the starting node. Default is True.

            Yields:
                source_tag, type_tag: str, str
        """
        debug_msg = f"[{self.__class__.__name__}] Descending {self.name} v{self.version} starting from ({source_tag},{type_tag})"
        logging.debug(debug_msg)

        source_gen = self.expand_tree(self.get_id(source_tag), mode=Tree.DEPTH)

        counter = 0
        for sid in source_gen:
            source_i = self.get_node(sid)
            types_i = source_i.data["sound_types"]

            if type_tag is None or type_tag == "%":
                if include_start_node or counter > 0:
                    yield source_i.tag, type_tag
                
                counter += 1

            else:
                if types_i.get_node(type_tag) is None:
                    debug_msg = f"[{self.__class__.__name__}] Sound source '{source_i.tag}' does not have sound type '{type_tag}'. Skipping ..."
                    logging.debug(debug_msg)
                    continue

                type_gen = types_i.expand_tree(types_i.get_id(type_tag), mode=Tree.DEPTH)
                for tid in type_gen:
                    type_i = types_i.get_node(tid)
                    if include_start_node or counter > 0:
                        yield source_i.tag, type_i.tag
                    
                    counter += 1


import os
import json
import copy
import warnings
from treelib import Tree


def tree_from_dict(tree, recipe, parent=None, data_transform=None):
    """ Transform a dictionary into a tree. 

        Within the dictionary, the key 'children' is used to designate branching 
        points. The key 'data' is used to designate any data associated with a 
        node.

        Args:
            tree: korus.tree.KTree
                Parent tree to which the data will be appended. Can be an empty tree.
            recipe: dict
                Dictionary recipe for building the tree.
            parent: str
                Tag or identifier of the node within the parent tree where the data will 
                be appended. By default parent=None implying that the data should be appended 
                at the root level.
            data_transform: callable
                This function gets applied to all entries in the dictionary with the 
                key 'data'.

        Returns:
            tree: Tree
                The created tree.
    """
    if data_transform is None:
        data_transform = lambda x: x

    branch_key = "children"

    for key,value in recipe.items():
        if "data" in value.keys():
            data = data_transform(value["data"])
        else:
            data = None

        nid = value["id"]
        tree.create_node(tag=key, identifier=nid, parent=parent, **data)            

        if branch_key in value.keys() and isinstance(value[branch_key], dict):
            tree = tree_from_dict(tree, value[branch_key], nid, data_transform)

    tree.clear_history()
    return tree


def tree_to_dict(tree, nid=None, key=None, sort=True, reverse=False):
    """ Transform a tree into a dictionary.
    
        The key 'children' is used to designate branching points.

        Args:
            tree: korus.tree.KTree
                Input tree.
            nid: str
                Only transform the part of the tree below this node.
                Default is root.
            key: str
                 @key and @reverse are present to sort nodes at each level.
                 If @key is None sorting is performed on node tag.
            sort: bool
                Whether to sort nodes. The default value is True.
            reverse: bool
                if True, reverse sorting

        Returns:
            tree_dict: dict()
                The created dictionary.
    """
    tree = Tree(tree, deep=True) #make a deep copy

    branch_key = "children"

    nid = tree.root if (nid is None) else nid
    ntag = tree[nid].tag
    tree_dict = {ntag: {branch_key: {}}}

    def parse_data_attr(x):
        if x is not None and isinstance(x, dict):
            for k,v in x.items():
                if isinstance(v, Tree):
                    x[k] = tree_to_dict(tree=v, sort=sort, reverse=reverse)
        return x

    tree_dict[ntag]["id"] = nid
    tree_dict[ntag]["data"] = parse_data_attr(tree[nid].data)

    if tree[nid].expanded:
        queue = [tree[i] for i in tree[nid].successors(tree._identifier)]
        key = (lambda x: x) if (key is None) else key

        if sort:
            queue.sort(key=key, reverse=reverse)

        for elem in queue:
            tree_dict[ntag][branch_key].update(tree_to_dict(tree, elem.identifier, sort=sort, reverse=reverse))

        if len(tree_dict[ntag][branch_key]) == 0:
            tree_dict = {ntag: {"id":nid, "data": parse_data_attr(tree[nid].data)}}

        return tree_dict


class KTree(Tree):
    """ KTree (Ketos-Tree) is derived from the treelib.tree.Tree 
        class. It adds a few new features and makes a few changes:

         * tags are required to be unique.
         * UUIDs are always used as identifiers.
         * sibling nodes can be merged with the `merge_nodes` method.
         * the node creation/removal history (precursor/heritor) is tracked.
         * a root node may be automatically created at construction time.
         * changes must be committed before certain operations are allowed, 
           e.g., removing or moving a newly created node.

        Args:
            root_tag: str
                Tag for the root node. If specified, the root node will be 
                automatically created at initialisation.
    """
    def __init__(self, root_tag="root"):
        super().__init__()
        
        self._tag_to_id = {}

        self._created_nodes = {}  # maps: created_node_uuid -> (precursor_node_uuid(s), is_equivalent)
        self._removed_nodes = {}  # maps: removed_node_uuid -> (inheritor_node_uuid(s), is_equivalent)

        self._data_merge_fcn = lambda x: dict()

        if root_tag is not None:
            self.create_node(root_tag)

    @property
    def created_nodes(self):
        return self._created_nodes

    @property
    def removed_nodes(self):
        return self._removed_nodes

    def deepcopy(self):
        """ Make a deep copy of the present instance
            See https://docs.python.org/2/library/copy.html
        """
        return copy.deepcopy(self)

    def clear_history(self):
        """ Clear the history of created and removed nodes.
        """
        self._created_nodes = {}
        self._removed_nodes = {}

    def show(self, append_name=False, **kwargs):
        """ Overwrites treelib.tree.Tree.show.
        
            Args:
                name: bool
                    Whether to append the node 'name' to the tag. Default is False.
        """
        if not append_name:
            print(self)
            ##super().show(**kwargs)
        
        else:
            tree = Tree(self, deep=True) #make a deep copy
            for n in tree.all_nodes_itr():
                if n.data is None:
                    continue

                if "name" in n.data:
                    n.tag += f' [{n.data["name"]}]'

            print(tree)
            ##tree.show(**kwargs)

    def _create_tag_to_id_mapping(self):
        """ Create tag -> identifier mapping 
        
            Returns:
                : dict
                    Dictionary mapping tags to identifiers
        """
        return {self.get_node(id).tag: id for id in self.expand_tree(mode=Tree.DEPTH)}            

    def get_node(self, n):
        """ Overwrites treelib.tree.Tree.get_node

            Args:
                n: str
                    Node tag or identifier
        """        
        return super().get_node(self.get_id(n))

    def rsearch(self, n, filter=None):
        """ Overwrites treelib.tree.Tree.rsearch

            Args:
                n: str
                    Node tag or identifier
            	filter: callable
                    Function of one variable to act on the Node object
        """        
        return super().rsearch(self.get_id(n), filter)

    def get_id(self, tag):
        """ Filters the tree nodes based on their tags.

            Note: If a valid identifier is specified as input, it will be returned unchanged.

            Args:
                tag: str,int,list,tuple
                    The selected tag(s) 
            
            Returns:
                ids: str,list
                    The corresponding node identifier(s).
        """
        if tag is None:
            return None
        
        is_list = isinstance(tag, (list,tuple))
        tag_list = tag if is_list else [tag]

        id_list = []
        for tag in tag_list:
            if tag in self._tag_to_id.keys(): #if input is a tag, map to id
                id_list.append(self._tag_to_id[tag]) 
            elif tag in self._tag_to_id.values(): #if input is already an id, do nothing
                id_list.append(tag) 
            else: #else, return None
                id_list.append(None)

        if is_list:
            return id_list
        else:
            return id_list[0]

    def create_node(self, tag, identifier=None, parent=None, precursor=None, **kwargs):
        """ Create a new, child node for a parent node.
        
            Node attributes can be specified using keyword arguments.

            Args:
                tag: str
                    Child node tag.
                identifier: str
                    Child node identifier. If absent, a UUID will be generated automatically.
                parent: str
                    Parent node identifier or tag. The default value is None, implying 'root' as parent.
                precursor: tuple
                    (IDs, is_equivalent) tuple, used for tracking the ancestry of the child node. 
                    If None, the parent identifier will be used.

            Returns: 
                node: treelib.node.Node
                    The new node object

            Raises:
                AssertionError: if the tree already contains a node with the specified tag
        """
        assert self.get_id(tag) is None, f"tree already contains a node with the tag '{tag}'"

        if kwargs is None or len(kwargs) == 0:
            kwargs = dict()

        parent = self.get_id(parent)
    
        node = super().create_node(tag=tag, identifier=identifier, parent=parent, data=kwargs)

        self._created_nodes[node.identifier] = ([self.get_id(parent)], False) if precursor is None else precursor
        self._tag_to_id[tag] = node.identifier

        return node

    def merge_nodes(self, tag, children=None, remove=False, data_merge_fcn=None, **kwargs):
        """ Create a new node by merging two or more existing nodes.

            Only sibling nodes can be merged in this manner.

            Use keyword arguments to specify attributes of the merged node.

            Args:
                tag: str
                    Tag of the new, merged node.
                children: list(str)
                    Identifiers or tags of the nodes to be merged. These will become the 
                    children nodes of the new, merged node.
                remove: bool
                    Remove source nodes after they have been merged. Default is False.
                data_merge_fcn: callable
                    Receives as input a list of the 'data' attributes of the source nodes 
                    and returns the 'data' attribute of the merged node.

            Returns: 
                node: treelib.node.Node
                    The new node object
                    
            Raises:
                AssertionError: if the tree already contains a node with the specified tag
                AssertionError: if the source nodes are not siblings
        """
        self._detect_new_node(children)

        if data_merge_fcn is None:
            data_merge_fcn = self._data_merge_fcn

        children = self.get_id(children)  # tag -> identifier

        # check that nodes are siblings (have the same parent)
        parent_nodes = [self.parent(id) for id in children]
        for parent_node in parent_nodes:
            assert parent_node.identifier == parent_nodes[0].identifier, "children must be siblings to merge"

        if kwargs is None or len(kwargs) == 0:
            kwargs = dict()

        # merge node data
        kwargs.update(data_merge_fcn([self.get_node(id).data for id in children]))

        # create new, merged node
        node = self.create_node(tag=tag, parent=parent_node.tag, precursor=(children, True), **kwargs)

        #move or remove original nodes
        for id in children:
            self.move_node(id, node.identifier) #move before removing, to ensure correct ancestry history
            if remove:
                self.remove_node(id)

        return node

    def remove_node(self, n):
        """ Overwrites treelib.tree.Tree.remove_node

            Args:
                n: str
                    Node tag or identifier
        """        
        self._detect_new_node(n)

        n = self.get_id(n) # tag -> identifier

        # set inheritors of removed nodes
        parent_id = self.parent(n).identifier
        node_gen = self.expand_tree(n, mode=Tree.DEPTH)
        for nid in node_gen:               
            self._removed_nodes[nid] = ([parent_id], False)  #(IDs, is_equivalent)

        return super().remove_node(n) # remove node and all nodes below it
    
    def move_node(self, n, new_parent):
        """ Overwrites treelib.tree.Tree.move_node
        
            Args:
                n: str
                    Node tag or identifier
        """
        self._detect_new_node(n)
        n = self.get_id(n) # tag -> identifier
        new_parent = self.get_id(new_parent)
        return super().move_node(n, new_parent)

    def link_past_node(self, n):
        """ Overwrites treelib.tree.Tree.link_past_node
        
            Args:
                n: str
                    Node tag or identifier
        """
        self._detect_new_node(n)
        n = self.get_id(n) # tag -> identifier

        # set inheritors of removed nodes
        self._removed_nodes[n] = ([child.identifier for child in self.children(n)], True)  #(IDs, is_equivalent)

        return super().link_past_node(n)

    def is_ancestor(self, ancestor, grandchild):
        """ Overwrites treelib.tree.Tree.is_ancestor
        
            Args:
                ancestor: str
                    Node tag or identifier
                grandchild: str
                    Node tag or identifier
        """
        ancestor = self.get_id(ancestor)
        grandchild = self.get_id(grandchild)
        return super().is_ancestor(ancestor, grandchild)

    def _detect_new_node(self, n):
        """ Check if a given node is new.

            Args:
                n: str, list(str)
                    Node tag or identifier
        
            Raises:
                RuntimeError: if any of the nodes are new
        """
        n_list = n if isinstance(n, (list,tuple)) else [n]
        for n in n_list:
            n = self.get_id(n)
            if n in self.created_nodes:
                raise RuntimeError("New node detected. Call clear_history before proceeding.")

    def last_common_ancestor(self, x):
        """ Finds the last common ancestor of a set of nodes
        
            Args:
                x: list(str)
                    List of node tags or identifiers

            Returns:
                : str
                    The tag of the last common ancestor, which may be one of the input nodes
        """
        nids = [self.get_id(n) for n in x]
        
        # find the shallowest node
        level_min = self.level(nids[0])
        nid_shallowest = nids[0]
        for nid in nids[1:]:            
            l = self.level(nid)
            if l < level_min:
                nid_shallowest = nid
                level_min = l

        # ascend the tree, at each step checking 
        # if we have found a common ancestor
        for nid_lca in self.rsearch(nid_shallowest):            
            is_lca = True
            for nid in nids:
                if not self.is_ancestor(nid_lca, nid) and nid_lca != nid:
                    is_lca = False
                    break    
        
            if is_lca:
                return self.get_node(nid_lca).tag



import os
import pytest
import json
import pandas as pd
from datetime import datetime, timedelta
import korus.tax as kx
import korus.db as kdb

path_to_assets = os.path.join(os.path.dirname(__file__), "assets")
path_to_tmp = os.path.join(path_to_assets, "tmp")


@pytest.fixture
def basic_db():
    sqlite_path = os.path.join(path_to_tmp, "basic_db.sqlite")
    if os.path.exists(sqlite_path):
        os.remove(sqlite_path)

    conn = kdb.create_db(sqlite_path)

    # create a fairly simple acoustic taxonomy
    tax = kx.AcousticTaxonomy(name="SimpleTax", path=sqlite_path)

    # add sources
    tax.create_sound_source("Bio", name="Biological sound source")
    tax.create_sound_source("Mammal", parent="Bio", name="Any mammal")
    kw = tax.create_sound_source("KW", parent="Mammal")
    kw = tax.create_sound_source("Dolphin", parent="Mammal")
    kw = tax.create_sound_source("HW", parent="Mammal")

    # add sound types
    tax.create_sound_type("TC", "Mammal", name="Tonal Call")
    tax.create_sound_type("CK", "Dolphin", name="Click")
    tax.create_sound_type("CK", "KW", name="Click")
    tax.create_sound_type("PC", "KW", name="Pulsed Call")
    tax.create_sound_type("W", "KW", name="Whistle")

    # save, version no. 1
    tax.save(comment="this is the first version")
    #print(tax)

    # add another source and some more sound types
    srkw = tax.create_sound_source("SRKW", parent="KW")
    tax.create_sound_type("S01", "SRKW", parent="PC", name="S1 call")
    tax.create_sound_type("S02", "SRKW", parent="PC", name="S2 call")

    srkw = tax.create_sound_source("NRKW", parent="KW")
    tax.create_sound_type("N01", "NRKW", parent="PC", name="N1 call")

    tax.save("added SRKW and NRKW") # version no. 2
    #print(tax)

    # remove NRKW again
    tax.remove_node("NRKW")
    tax.save("removed NRKW") # version no. 3
    #print(tax)

    # link past KW
    tax.link_past_node("KW")
    tax.save("linked past KW") # version no. 4
    #print(tax)

    yield conn, sqlite_path

    conn.close()
    if os.path.exists(sqlite_path):
        os.remove(sqlite_path)


@pytest.fixture
def deploy_data():
    lat = 49.780487
    lon = -122.05154
    depth = 18.0
    v = {
        "owner": "OceanResearch",
        "name": "WestPoint",
        "start_utc": "2022-06-24",
        "end_utc": "2022-10-03",
        "location": "Salt Spring Island, BC, Canada",
        "latitude_deg": lat,
        "longitude_deg": lon,
        "depth_m": depth,
        "latitude_min_deg": lat,
        "latitude_max_deg": lat,
        "longitude_min_deg": lon,
        "longitude_max_deg": lon,
        "depth_min_m": depth,
        "depth_max_m": depth,
        "license": None,
        "hydrophone": None,
        "bits_per_sample": None,
        "sample_rate": 128000,
        "num_channels": 1,
        "sensitivity": None,
        "comments": None
    }
    return v


@pytest.fixture
def file_data():

    def abclisten_timestamp_parser(x):
        fmt = "_%Y%m%dT%H%M%S.%fZ"
        p = x.find("_")
        s = x[p: p + 21]
        return datetime.strptime(s, fmt)

    fnames =[
        "ABCLISTENHF1234_20220624T164000.000Z_20220624T164459.996Z.flac",
        "ABCLISTENHF1234_20220624T164500.023Z_20220624T164959.994Z.flac"
    ]

    deploy_path = "OceanResearch/WestPoint"

    file_data = []
    for fname in fnames:
        dt = abclisten_timestamp_parser(fname)
        dir_path = os.path.join(deploy_path, dt.strftime("%Y%m%d"))
        num_samples, sample_rate = 32000*5*60, 32000
        start_utc_str = dt.strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        end_utc_str = (dt+timedelta(seconds=num_samples/sample_rate)).strftime("%Y-%m-%d %H:%M:%S.%f")[:-3]
        v = {
            "deployment_id": 1,
            "storage_id": 1,
            "filename": fname,
            "relative_path": dir_path,
            "sample_rate": sample_rate,
            "downsample": None,
            "num_samples": num_samples,
            "format": "FLAC",
            "codec": "FLAC",
            "start_utc": start_utc_str,
            "end_utc": end_utc_str,
        }
        file_data.append(v)

    return file_data


@pytest.fixture
def db_with_annotations(basic_db, deploy_data, file_data):
    """ Creates a database with a small set of annotations.

        Database contents:

            * A single hydrophone deployment
            * Two, consecutive 5-minute audio files sampled at 32 kHz, with the first audio 
                file starting at 2022-06-24 16:40:00.000
            * An exhaustive annotation job targeting KW,PC and KW,W
            * Three annotations:
                - KW,PC: starting 30.0s into the first audio file and lasting 1.3s
                - SRKW,S01: starting 21.1s into the first audio file and lasting 5 minutes
                - None,None: a manual negative starting 1.0s into the first audio file and lasting 0.8s
                - auto-generated negatives
    """
    (conn, sqlite_path) = basic_db

    c = conn.cursor()
    c = kdb.insert_row(conn, table_name="deployment", values=deploy_data)

    v = {
        "name": "laptop",
        "path": "/",
        "description": "data from my latest deployments",
    }
    c = kdb.insert_row(conn, table_name="storage", values=v)

    for v in file_data:
        kdb.insert_row(conn, table_name="file", values=v)

    # insert job
    primary_sound = kdb.get_label_id(
        conn, 
        source_type=[("KW","PC"), ("KW","W")],
        taxonomy_id=2,
        always_list=True
    )
    background_sound = kdb.get_label_id(
        conn, 
        source_type=("%","CK"),
        taxonomy_id=2,
        always_list=True
    )
    v = {
        "taxonomy_id": 2,
        "model_id": None,
        "annotator": "LL",
        "primary_sound": json.dumps(primary_sound),
        "background_sound": json.dumps(background_sound),
        "is_exhaustive": 1,
        "configuration": None,
        "start_utc": "2022-10",
        "end_utc": "2023-03",
        "by_human": 1,
        "by_machine": 0,
        "comments": "Vessel noise annotated opportunistically",
        "issues": json.dumps(["start and end times may not always be accurate", "some KW sounds may have been incorrectly labelled as HW"]),
    }
    c = kdb.insert_row(conn, table_name="job", values=v)

    # link files
    # get deployment id
    query = """
        SELECT 
            id
        FROM 
            deployment 
        WHERE 
            owner LIKE 'OceanResearch' 
            AND name LIKE 'WestPoint'
            AND start_utc >= '2022-01-01'
            AND end_utc <= '2022-12-31'
    """
    rows = c.execute(query).fetchall()
    deploy_id = rows[0][0]

    # get file_id
    for v in file_data:
        fname = v["filename"]

        query = f"""
            SELECT 
                id
            FROM 
                file 
            WHERE 
                deployment_id = '{deploy_id}' 
                AND filename LIKE '{fname}'
        """
        rows = c.execute(query).fetchall()
        file_id = rows[0][0]

        # link file to job
        v = {
            "job_id": 1,
            "file_id": file_id,
            "channel": 0
        }
        c = kdb.insert_row(conn, table_name="file_job_relation", values=v)

    # define a tag
    v = {
        "name": "NEGATIVE",
        "description": "A negative sample"
    }
    c = kdb.insert_row(conn, table_name="tag", values=v)

    #insert annotation using add_annotation function
    annot_tbl = pd.DataFrame({
        "file_id": [1,1,1],
        "channel": [0,0,0],
        "sound_source": ["KW","SRKW",None],
        "sound_type": ["PC","S01",None],
        "tentative_sound_source": ["SRKW",None,None],
        "tentative_sound_type": ["S01",None,None],
        "tag": [None,None,["NEGATIVE"]],
        "duration_ms": [1300,300000,800],
        "start_ms": [30000,21200,1000],
        "freq_min_hz": [600,700,None],
        "freq_max_hz": [4400,3300,None],
        "granularity": ["unit","window","window"],
        "comments": ["no additional observations","","this is a negative sample"]
    })
    annot_ids = kdb.add_annotations(conn, annot_tbl=annot_tbl, job_id=1)
    neg_ids = kdb.add_negatives(conn, job_id=1)

    conn.commit()

    yield conn, sqlite_path

    conn.close()
    if os.path.exists(sqlite_path):
        os.remove(sqlite_path)



import os
import pytest
import subprocess

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(current_dir, "assets")
path_to_tmp = os.path.join(path_to_assets, "tmp")


def test_tutorials():
    """ Check that all Jupyter Notebook tutorials complete successfully 
    and generate the expected output files (we only check that the files 
    exist, not that they have the expected content)"""
    bash_script_path = os.path.join(current_dir, "run_tutorials.sh")
    res = subprocess.check_output([bash_script_path]).decode()
    assert res == "tax_t1.sqlite created\nmydb.sqlite created\n"



def create_annotation_table(conn):
    """ Create annotation table according to Korus schema.

        TODO: Change tentative_label_id type from INTEGER to JSON ? (to allow for lists)
                Or add another column named label_list_id (or similar)
    
        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()

    tbl_def = """
        CREATE TABLE
            annotation(
                id INTEGER NOT NULL,
                job_id INTEGER NOT NULL,
                deployment_id INTEGER NOT NULL,
                file_id INTEGER,
                label_id INTEGER,
                tentative_label_id INTEGER,
                ambiguous_label_id JSON,
                excluded_label_id JSON,
                tag_id JSON, 
                granularity_id INTEGER NOT NULL DEFAULT 2,
                num_files INTEGER NOT NULL DEFAULT 1,
                file_id_list JSON NOT NULL,
                start_utc TEXT,
                duration_ms INTEGER,
                start_ms INTEGER DEFAULT 0,
                freq_min_hz INTEGER DEFAULT 0,
                freq_max_hz INTEGER,
                channel INTEGER NOT NULL DEFAULT 0,
                machine_prediction JSON,
                valid INTEGER NOT NULL DEFAULT 1,
                comments TEXT,
                PRIMARY KEY (id),
                FOREIGN KEY (label_id) REFERENCES label (id),
                FOREIGN KEY (tentative_label_id) REFERENCES label (id),
                FOREIGN KEY (job_id) REFERENCES job (id),
                FOREIGN KEY (file_id) REFERENCES file (id),
                FOREIGN KEY (deployment_id) REFERENCES deployment (id),
                FOREIGN KEY (granularity_id) REFERENCES granularity (id),
                CHECK (
                    duration_ms > 0
                    AND freq_min_hz < freq_max_hz
                )
            )
        """
    c.execute(tbl_def)


def create_granularity_table(conn):
    """ Create granularity table according to Korus schema.

        Also adds entries for the standard Korus granularities: unit, window, file, batch, encounter

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()
    tbl_def = """
        CREATE TABLE
            granularity(
                id INTEGER NOT NULL,
                name TEXT NOT NULL,
                description TEXT,
                PRIMARY KEY (id),
                UNIQUE (name)
            )
        """
    c.execute(tbl_def)

    rows = [
        (
            "unit",
            "Annotation of a single vocalisation/sound. Bounding box drawn snuggly around a single vocalisation/sound."\
            " Overlapping sounds may be present."
        ),
        (
            "window",
            "Annotation of a single vocalisation/sound. Box width does not necessarily match sound duration."\
            " Sound may be fully or only partially within the box. Overlapping sounds may be present."
        ),
        (
            "file",
            "Annotation spanning precisely the duration of a single audio file."
        ),
        (
            "batch",
            "Annotation of multiple vocalisations/sounds."
        ),
        (
            "encounter",
            "Annotation of an entire (biological) acoustic encounter."
        ),
    ]

    for row in rows:
        name = row[0]
        descr = row[1]
        c.execute(
            f"INSERT INTO granularity (id,name,description) VALUES (NULL,?,?)", [name,descr]
        )


def create_job_table(conn):
    """ Create job table according to Korus schema.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()

    tbl_def = """
        CREATE TABLE
            job(
                id INTEGER NOT NULL,
                taxonomy_id INTEGER,
                model_id INTEGER,
                annotator TEXT,
                primary_sound JSON,
                background_sound JSON,
                is_exhaustive INTEGER,
                configuration JSON,
                start_utc TEXT,
                end_utc TEXT,
                by_human INTEGER NOT NULL DEFAULT 1,
                by_machine INTEGER NOT NULL DEFAULT 0,
                issues JSON,
                comments TEXT,
                PRIMARY KEY (id),
                FOREIGN KEY (taxonomy_id) REFERENCES taxonomy (id),
                FOREIGN KEY (model_id) REFERENCES model (id),
                CHECK (
                    is_exhaustive IN (0, 1)
                    AND by_human IN (0, 1)
                    AND by_machine IN (0, 1)
                    AND (by_human > 0 OR by_machine > 0)
                    AND start_utc < end_utc
                )
            )
        """
    c.execute(tbl_def)


def create_deployment_table(conn):
    """ Create deployment table according to Korus schema.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()

    tbl_def = """
        CREATE TABLE
            deployment(
                id INTEGER NOT NULL,
                name TEXT NOT NULL,
                owner TEXT,
                start_utc TEXT,
                end_utc TEXT,
                location TEXT,
                latitude_deg REAL,
                longitude_deg REAL,
                depth_m REAL,
                trajectory JSON,
                latitude_min_deg REAL,
                latitude_max_deg REAL,
                longitude_min_deg REAL,
                longitude_max_deg REAL,
                depth_min_m REAL,
                depth_max_m REAL,
                license TEXT,
                hydrophone TEXT,
                bits_per_sample INTEGER,
                sample_rate INTEGER,
                num_channels INTEGER,
                sensitivity REAL,
                comments TEXT,
                PRIMARY KEY (id),
                UNIQUE (owner, name, start_utc, end_utc),
                CHECK (
                    latitude_deg BETWEEN -90 AND 90
                    AND longitude_deg BETWEEN -180 AND 180
                    AND depth_m BETWEEN 0 and 11000
                    AND start_utc <= end_utc 
                )
            )
        """
    c.execute(tbl_def)


def create_file_table(conn):
    """ Create file table according to Korus schema.

        Also creates an index on (deployment_id, filename) for faster querying.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()

    # create table
    tbl_def = """
        CREATE TABLE
            file(
                id INTEGER NOT NULL,
                deployment_id INTEGER NOT NULL,
                storage_id INTEGER NOT NULL,
                filename TEXT NOT NULL,
                relative_path TEXT NOT NULL,
                sample_rate INTEGER NOT NULL,
                num_samples INTEGER NOT NULL,
                downsample TEXT,
                format TEXT,
                codec TEXT,
                start_utc TEXT,
                end_utc TEXT,
                PRIMARY KEY (id),
                FOREIGN KEY (deployment_id) REFERENCES deployment (id),
                FOREIGN KEY (storage_id) REFERENCES storage (id),
                UNIQUE (deployment_id, filename, relative_path)
            )
        """
    c.execute(tbl_def)

    # create indices for faster queries
    c.execute("""
        CREATE INDEX
            deployment_filename_index
        ON
            file(deployment_id, filename)
    """
    )

    c.execute("""
        CREATE INDEX
            deployment_time_index
        ON
            file(deployment_id, start_utc, end_utc)
    """
    )


def create_file_job_relation_table(conn):
    """ Create file-job relation table according to Korus schema.

        Also creates an index on (job_id) for faster querying.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()

    # create table
    tbl_def = """
        CREATE TABLE
            file_job_relation(
                job_id INTEGER NOT NULL,
                file_id INTEGER NOT NULL,
                channel INTEGER NOT NULL DEFAULT 0,
                FOREIGN KEY (job_id) REFERENCES job (id),
                FOREIGN KEY (file_id) REFERENCES file (id),
                UNIQUE (job_id, file_id, channel)
            )
        """
    c.execute(tbl_def)

    # create index for faster queries
    c.execute("""
        CREATE INDEX
            job_index
        ON
            file_job_relation(job_id)
    """
    )


def create_model_table(conn):
    """ Create model table according to Korus schema.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()
    tbl_def = """
        CREATE TABLE
            model(
                id INTEGER NOT NULL,
                name TEXT NOT NULL,
                version TEXT,
                data JSON NOT NULL,
                PRIMARY KEY (id),
                UNIQUE (name, version)
            )
        """
    c.execute(tbl_def)


def create_storage_table(conn):
    """ Create data-storage table according to Korus schema.

        @address can be an IP address or a URL

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()
    tbl_def = """
        CREATE TABLE
            storage(
                id INTEGER NOT NULL,
                name TEXT NOT NULL,
                path TEXT NOT NULL DEFAULT '/',
                address TEXT,
                description TEXT,
                PRIMARY KEY (id),
                UNIQUE (name, path, address)
            )
        """
    c.execute(tbl_def)


AUTO_NEG = "__AUTO_NEGATIVE__"

def create_tag_table(conn):
    """ Create tag table according to Korus schema.

        Also adds an entry for auto-generated negatives.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()
    tbl_def = """
        CREATE TABLE
            tag(
                id INTEGER NOT NULL,
                name TEXT NOT NULL,
                description TEXT,
                PRIMARY KEY (id),
                UNIQUE (name)
            )
        """
    c.execute(tbl_def)

    # add entry for AUTO_NEG
    descr = "Negative sample, automatically generated by the korus.db.add_negatives function"
    c.execute(
        f"INSERT INTO tag (id,name,description) VALUES (NULL,?,?)", [AUTO_NEG,descr]
    )


def create_taxonomy_table(conn):
    """ Create taxonomy table according to Korus schema.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()
    tbl_def = """
        CREATE TABLE
            taxonomy(
                id INTEGER NOT NULL,
                name TEXT NOT NULL,
                version INTEGER,
                tree JSON NOT NULL,
                timestamp TEXT,
                comment TEXT,
                PRIMARY KEY (id),
                UNIQUE (name, version)
            )
        """
    c.execute(tbl_def)


def create_taxonomy_created_node_table(conn):
    """ Create taxonomy_created_node table according to Korus schema.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()
    tbl_def = """
        CREATE TABLE
            taxonomy_created_node(
                id TEXT NOT NULL,
                precursor_id JSON NOT NULL,
                is_equivalent INTEGER NOT NULL,
                taxonomy_id INTEGER NOT NULL,
                PRIMARY KEY (id),
                FOREIGN KEY (taxonomy_id) REFERENCES taxonomy (id)
            )
        """
    c.execute(tbl_def)


def create_taxonomy_removed_node_table(conn):
    """ Create taxonomy_removed_node table according to Korus schema.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()
    tbl_def = """
        CREATE TABLE
            taxonomy_removed_node(
                id TEXT NOT NULL,
                inheritor_id JSON NOT NULL,
                is_equivalent INTEGER NOT NULL,
                taxonomy_id INTEGER NOT NULL,
                PRIMARY KEY (id),
                FOREIGN KEY (taxonomy_id) REFERENCES taxonomy (id)
            )
        """
    c.execute(tbl_def)


def create_label_table(conn):
    """ Create label table according to Korus schema.

        Also creates an index on (taxonomy_id, sound_source_tag, sound_type_tag) for faster querying.

        Args:
            conn: sqlite3.Connection
                Database connection
    """
    c = conn.cursor()

    # create table
    tbl_def = """
        CREATE TABLE
            label(
                id INTEGER NOT NULL,
                taxonomy_id INTEGER NOT NULL,
                sound_source_tag TEXT,
                sound_source_id TEXT,
                sound_type_tag TEXT,
                sound_type_id TEXT,
                PRIMARY KEY (id),
                FOREIGN KEY (taxonomy_id) REFERENCES taxonomy (id),
                UNIQUE (taxonomy_id, sound_source_id, sound_type_id)
            )
        """
    c.execute(tbl_def)

    # create index for faster queries
    c.execute("""
        CREATE INDEX
            source_type_index
        ON
            label(taxonomy_id, sound_source_tag, sound_type_tag)
    """
    )
        
    #c.execute("INSERT INTO label VALUES (NULL, NULL, NULL)")



import os
import numpy as np
import pytest
import sqlite3
import korus.db as kdb
import korus.selection as ks

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(current_dir, "assets")
path_to_tmp = os.path.join(path_to_assets, "tmp")

np.random.seed(1)


def test_create_selections(db_with_annotations):
    conn, db_path = db_with_annotations
    idx_kw = kdb.filter_annotation(conn, source_type=("KW","PC"), taxonomy_id=3)
    assert len(idx_kw) == 2

    # create KW selections without stepping
    df = ks.create_selections(conn, indices=idx_kw, window_ms=3000,)
    assert len(df) == 2
    assert np.all(df.columns.values == ["start", "end", "annot_id"])
    assert np.all(df.index.names == ["sel_id", "filename"])
    assert np.all(np.isclose((df.end - df.start).values, 3.0, rtol=1e-3))

    # create KW selections with stepping
    df = ks.create_selections(conn, indices=idx_kw, window_ms=3000, step_ms=1000, full_path=False)

    # two selections created from first annotation (1.3s duration)
    df_1 = df[df.annot_id == 1]
    assert len(df_1) == 2

    # about 300 selections created from second annotation (300s duration)
    df_2 = df[df.annot_id == 2]
    assert len(df_2) == 300

    # create KW selections with stepping and cap on max number of selections
    df = ks.create_selections(conn, indices=idx_kw, window_ms=3000, step_ms=1000, num_max=10)
    df_1 = df[df.annot_id == 0]
    assert len(df_1) == 0
    df_2 = df[df.annot_id == 2]
    assert len(df_2) == 10

    # create KW selections with exclusive=True and window=1min and step=50s
    df = ks.create_selections(conn, indices=idx_kw, window_ms=60000, step_ms=50000, exclusive=True)
    df_1 = df[df.annot_id == 0]
    assert len(df_1) == 0
    df_2 = df[df.annot_id == 2]
    assert len(df_2.index.get_level_values(0).unique()) == 5
    assert df.iloc[0].start > 21.2
    assert df.iloc[-1].end < 21.2
    




import os
import pytest
from datetime import datetime
import pandas as pd
import korus.util as ku


# TODO: add tests for all functions in korus.util module


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(current_dir, "assets")
path_to_tmp = os.path.join(path_to_assets, "tmp")


def test_find_files_dir():
    """Check that we can find files in a directory"""
    path = os.path.join(path_to_assets, "files")
    f = ku.find_files(path)
    f.sort()
    assert f == ["a.txt", "b.wav"]
    f = ku.find_files(path, subdirs=True)
    f.sort()
    assert f == ["a.txt", "b.wav", "more-files/c.flac", "more-files/d.wav", "more-files/e.txt"]
    f = ku.find_files(path, substr="wav", subdirs=True)
    f.sort()
    assert f == ["b.wav", "more-files/d.wav"]

def test_find_files_tar():
    """Check that we can find files in a zipped tar archive"""
    path = os.path.join(path_to_assets, "zipped-files.tar.gz")
    f = ku.find_files(path, subdirs=True)
    f.sort()
    assert f == ["a.txt", "b.wav", "more-files/c.flac", "more-files/d.wav", "more-files/e.txt"]
    f = ku.find_files(path, substr="wav")
    f.sort()
    assert f == ["b.wav"]
    f = ku.find_files(path, substr="wav", subdirs=True)
    f.sort()
    assert f == ["b.wav", "more-files/d.wav"]
    f = ku.find_files(path, substr="wav", tar_path="/more-files")
    f.sort()
    assert f == ["d.wav"]
    f = ku.find_files(path, substr="wav", tar_path="non-existing-path")
    f.sort()
    assert f == []

@pytest.mark.parametrize("rel_path", ["timestamped-audiofiles", "zipped-timestamped-audiofiles.tar.gz"])
def test_collect_audiofile_metadata(rel_path):
    """Check that we can extract metadata from audio files both stored 
        within a regular directory and within a zipped tar archive"""
    def timestamp_parser(x):
        return datetime.strptime(x[9:28], "%Y%m%dT%H%M%S.%f")
        
    path = os.path.join(path_to_assets, rel_path)

    # by default, only search for WAV files
    df = ku.collect_audiofile_metadata(path)
    assert len(df) == 0

    # search for FLAC files
    df = ku.collect_audiofile_metadata(path, ext="FLAC")
    # compare to expected result
    answ_path = os.path.join(path_to_assets, "timestamped-audiofiles/df-no-timestamps.csv")
    answ = pd.read_csv(answ_path).astype({"relative_path": "str"})
    pd.testing.assert_frame_equal(df, answ)

    # search for FLAC files with timestamp parser
    df = ku.collect_audiofile_metadata(path, ext="FLAC", timestamp_parser=timestamp_parser)
    # compare to expected result
    answ_path = os.path.join(path_to_assets, "timestamped-audiofiles/df-timestamps.csv")
    answ = pd.read_csv(answ_path).astype({"relative_path": "str"})
    pd.testing.assert_frame_equal(df, answ)

    # search with time constraints and date_subfolder=True
    earliest_start_utc = datetime(2024,6,30,12,0,0)
    df = ku.collect_audiofile_metadata(
        path, 
        ext="FLAC", 
        timestamp_parser=timestamp_parser,
        earliest_start_utc=earliest_start_utc,
        date_subfolder=True,    
    )
    # compare to expected result
    answ_2024 = answ[answ.start_utc.apply(lambda x: "2024" in x)].reset_index(drop=True)
    pd.testing.assert_frame_equal(df, answ_2024)


def test_collect_audiofile_metadata_from_tar():
    """Check that we can extract metadata from audio files within a zipped tar archive"""
    pass


import json
import korus.db as kdb
from korus.util import list_to_str


def get_related_label_id(conn, label_id, ascend=False, descend=False, always_list=False):
    """ Get label identifiers of related nodes in the taxonomy tree.

        Args:
            conn: sqlite3.Connection
                Database connection
            label_id: int, list(int)
                Label identifiers
            ascend: bool
                Return the labels of ancestral nodes. 
            descend: bool
                Return the labels of descendant nodes. 
            always_list: bool
                Whether to always return a list. Default is False.
            
        Returns:
            related_id: int
                Identifiers of related nodes        
    """
    c = conn.cursor()

    related_id = []
    if not ascend and not descend:
        return related_id

    #loop over input label identifiers
    for _label_id in label_id if isinstance(label_id, list) else [label_id]: 

        # get source/type tags and taxonomy id
        query = f"SELECT sound_source_tag, sound_type_tag, taxonomy_id FROM label WHERE id = {_label_id}"
        (sound_source, sound_type, tax_id) = c.execute(query).fetchall()[0]          

        tax = kdb.get_taxonomy(conn, tax_id)

        if ascend:
            gen = tax.ascend(sound_source, sound_type, include_start_node=False)
            for i,(s,t) in enumerate(gen): #generator returns (source,type) tag tuple
                query = f"""
                    SELECT 
                        id 
                    FROM 
                        label 
                    WHERE 
                        sound_source_tag = '{s}'
                        AND sound_type_tag = '{t}'
                        AND taxonomy_id = {tax_id}
                    """
                (id,) = c.execute(query).fetchall()[0]
                related_id.append(id)

        if descend:
            gen = tax.descend(sound_source, sound_type, include_start_node=False)
            for i,(s,t) in enumerate(gen): #generator returns (source,type) tag tuple
                query = f"""
                    SELECT 
                        id 
                    FROM 
                        label 
                    WHERE 
                        sound_source_tag = '{s}'
                        AND sound_type_tag = '{t}'
                        AND taxonomy_id = {tax_id}
                    """
                (id,) = c.execute(query).fetchall()[0]
                related_id.append(id)

    if len(related_id) == 1 and not always_list:
        related_id = related_id[0]

    return related_id


def crosswalk_label_ids(conn, label_id, dst_taxonomy_id, ascend=False, descend=False, always_list=False, equiv=False):
    """ Map a list of label identifiers to another taxonomy.

        Same as @crosswalk_label_id, but for list inputs.

        Args:
            conn: sqlite3.Connection
                Database connection
            label_id: list(int)
                Label identifier
            dst_taxonomy_id: int
                Destination taxonomy identifier
            ascend: bool
                Also return the labels of ancestral nodes of the mapped node(s). 
            descend: bool
                Also return the labels of descendant nodes of the mapped node(s). 
            equiv: bool
                If True, only return the mapped label IDs that are 1-to-1

        Returns:
            mapped_label_id: list(str)
                The mapped label identifier(s).
    """
    mapped_label_ids = []
    for id in label_id:
        mapped_id, is_equiv = crosswalk_label_id(
            conn=conn, 
            label_id=id, 
            dst_taxonomy_id=dst_taxonomy_id, 
            ascend=ascend, 
            descend=descend, 
            always_list=True
        )
        if (equiv and is_equiv) or not equiv:
            mapped_label_ids += mapped_id

    return mapped_label_ids

def crosswalk_label_id(conn, label_id, dst_taxonomy_id, ascend=False, descend=False, always_list=False):
    """ Map a single label identifier to another taxonomy.

        Args:
            conn: sqlite3.Connection
                Database connection
            label_id: int
                Label identifier
            dst_taxonomy_id: int
                Destination taxonomy identifier
            ascend: bool
                Also return the labels of ancestral nodes of the mapped node(s). 
            descend: bool
                Also return the labels of descendant nodes of the mapped node(s). 
            always_list: bool
                Whether to always return a list. Default is False.

        Returns:
            mapped_label_id: list(str)
                The mapped label identifier(s).
            is_equivalent: bool
                Whether the input label and the mapped label(s) may be considered equivalent. 
    """
    c = conn.cursor()

    # special case: (None, None)
    query = f"SELECT sound_source_tag,sound_type_tag FROM label WHERE id = {label_id}"
    (ss, st) = c.execute(query).fetchall()[0]
    if ss is None and st is None:
        query = f"""
            SELECT 
                id 
            FROM 
                label 
            WHERE 
                sound_source_tag IS NULL 
                AND sound_type_tag IS NULL 
                AND taxonomy_id = {dst_taxonomy_id}
        """
        (mapped_label_id,) = c.execute(query).fetchall()[0]
        if always_list:
            mapped_label_id = [mapped_label_id]

        return mapped_label_id, True


    # origin taxonomy id
    query = f"SELECT taxonomy_id FROM label WHERE id = '{label_id}'"
    rows = c.execute(query).fetchall()
    (org,) = rows[0]

    # destination taxonomy
    dst = dst_taxonomy_id

    if org == dst:
        mapped_label_id = [label_id]
        mapped_label_id += get_related_label_id(conn, mapped_label_id, descend=descend, ascend=ascend, always_list=True)
        if not always_list and len(mapped_label_id) == 1:
            mapped_label_id = mapped_label_id[0]
        return mapped_label_id, True
        
    # sound source and sound type UUIDs in origin taxonomy
    query = f"SELECT sound_source_id, sound_type_id FROM label WHERE id = '{label_id}'"
    rows = c.execute(query).fetchall()
    source_id, type_id = rows[0]

    # check if the same UUID tuple exists in the dst taxonomy
    query = f"""
        SELECT 
            id
        FROM 
            label 
        WHERE 
            sound_source_id = '{source_id}' 
            AND sound_type_id = '{type_id}' 
            AND taxonomy_id = '{dst}'
        """
    rows = c.execute(query).fetchall()
    mapped_label_id = [row[0] for row in rows]
    if len(rows) == 1:
        mapped_label_id += get_related_label_id(conn, mapped_label_id, descend=descend, ascend=ascend, always_list=True)
        if not always_list:
            mapped_label_id = mapped_label_id[0]
        return mapped_label_id, True
    
    # forward or backward mode
    mode = "b" if dst < org else "f"

    # trace node history
    mapped_source_id, source_is_equivalent = trace_node_history(conn, source_id, dst, mode)
    mapped_type_id, type_is_equivalent = trace_node_history(conn, type_id, dst, mode)

    is_equivalent = source_is_equivalent and type_is_equivalent

    # finally, query label identifier using the mapped UUIDs
    mapped_source_id_str = list_to_str(mapped_source_id)  
    mapped_type_id_str = list_to_str(mapped_type_id) 
    query = f"""
        SELECT 
            id
        FROM 
            label 
        WHERE 
            sound_source_id IN {mapped_source_id_str} 
            AND sound_type_id IN {mapped_type_id_str}  
            AND taxonomy_id = '{dst}'
        """
    rows = c.execute(query).fetchall()
    mapped_label_id = [row[0] for row in rows]
    mapped_label_id += get_related_label_id(conn, mapped_label_id, descend=descend, ascend=ascend, always_list=True)

    if not always_list and len(mapped_label_id) == 1:
        mapped_label_id = mapped_label_id[0]

    return mapped_label_id, is_equivalent


def crosswalk_label(conn, source_type, origin_taxonomy_id, dst_taxonomy_id, ascend=False, descend=False, always_list=False):
    """ Map a label (sound-source, sound-type tag tuple) to another taxonomy.

        Wrapper for :func:`crosswalk_label_id`

        Args:
            conn: sqlite3.Connection
                Database connection
            source_type: tuple(str, str)
                Sound-source, sound-type tag tuple. 
            origin_taxonomy_id: int
                Origin taxonomy identifier
            dst_taxonomy_id: int
                Destination taxonomy identifier
            ascend: bool
                Also return the labels of ancestral nodes of the mapped node(s). 
            descend: bool
                Also return the labels of descendant nodes of the mapped node(s). 
            always_list: bool
                Whether to always return a list. Default is False.

        Returns:
            mapped_label: list(tuple)
                The mapped label(s).
            is_equivalent: bool
                Whether the input label and the mapped label(s) may be considered equivalent. 
    """
    c = conn.cursor()

    org = origin_taxonomy_id
    dst = dst_taxonomy_id

    # map tag tuple -> label_id in origin taxonomy
    (source_tag, type_tag) = source_type
    query = f"""
        SELECT 
            id
        FROM 
            label 
        WHERE 
            sound_source_tag = '{source_tag}' 
            AND sound_type_tag = '{type_tag}' 
            AND taxonomy_id = '{org}'
        """
    rows = c.execute(query).fetchall()
    (label_id, ) = rows[0]

    mapped_label_id, is_equivalent = crosswalk_label_id(
        conn, 
        label_id, 
        dst_taxonomy_id=dst, 
        always_list=always_list,
        ascend=ascend,
        descend=descend
    )

    # map label_id -> tag tuple
    mapped_label_id_str = list_to_str(mapped_label_id)
    query = f"""
        SELECT 
            sound_source_tag, sound_type_tag
        FROM 
            label 
        WHERE 
            id IN {mapped_label_id_str} 
        """
    mapped_label = c.execute(query).fetchall()

    return mapped_label, is_equivalent


def trace_node_history(conn, node_id, dst_taxonomy_id, mode="backward"):
    """ Maps a taxonomy node to an earlier taxonomy version. 

        Args:
            conn: sqlite3.Connection
                Database connection
            node_id: str
                Sound source or sound type UUID.
            dst_taxonomy_id: int
                Destination taxonomy identifier
            mode: str
                * backward/b: trace node history backwards in time (default)
                * forward/f: trace node history forward in time

        Returns:
            mapped_ids: list(str)
                UUID(s) in the destination taxonomy.
            is_equivalent: bool
                Whether the input node and the mapped node(s) may be considered equivalent.             
    """
    mode = mode.lower()[0]
    assert mode == "b" or mode == "f", "mode must be either backward (b) or forward (f)"

    if mode == "b":
        col_name = "precursor_id"
        table_name = "taxonomy_created_node" 
    else:
        col_name = "inheritor_id"
        table_name = "taxonomy_removed_node"

    dst = dst_taxonomy_id
    ids = [node_id]    
    mapped_ids = []

    c = conn.cursor()

    is_equivalent = True

    while True:
        for id in ids:
            query = f"""
                SELECT 
                    id 
                FROM 
                    label 
                WHERE 
                    (sound_source_id = '{id}' OR sound_type_id = '{id}') 
                    AND taxonomy_id = '{dst}'
            """
            rows = c.execute(query).fetchall()
            if len(rows) > 0:
                mapped_ids.append(id)

        ids = [x for x in ids if x not in mapped_ids]

        if len(ids) == 0:
            break

        ids_str = list_to_str(ids)  #"(" + ",".join([f"'{x}'" for x in ids]) + ")"
        query = f"""
            SELECT 
                {col_name}, is_equivalent
            FROM 
                {table_name}
            WHERE 
                id IN {ids_str}
            """
        rows = c.execute(query).fetchall()

        ids = []
        for row in rows:
            (precursor_id, equiv) = row
            ids += json.loads(precursor_id)
            if not equiv:
                is_equivalent = False

    return mapped_ids, is_equivalent



from setuptools import setup, find_packages
from korus.__init__ import __version__

setup(
    name="korus",
    version=__version__,
    description="Python package for managing acoustic metadata and annotations",
    author="Oliver Kirsebom",
    author_email="oliver.kirsebom@gmail.com",
    url='https://github.com/meridian-analytics/korus',
    license="",
    packages=find_packages(),
    install_requires=[
        "soundfile", 
        "pandas", 
        "numpy",
        "termcolor", 
        "treelib", 
        "tqdm",
        "tabulate",
        "pyyaml",
    ],
    entry_points={
        "console_scripts": [
            "korus-submit = korus.app.submit:main",
        ],
    },
    setup_requires=["pytest-runner", "wheel"],
    tests_require=["pytest",],
    include_package_data=True,
    zip_safe=False,
)



import os
import pytest
from io import StringIO
import pandas as pd
import numpy as np
from datetime import datetime, timedelta
import korus.db_util.label as klb
import json

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(current_dir, "assets")
path_to_tmp = os.path.join(path_to_assets, "tmp")


def test_crosswalk_label(basic_db):
    """Test that crosswalk_label function can be used to map labels across taxonomy versions"""
    (conn, path) = basic_db

    label_org = ("KW","PC")
    label_dst, is_equiv = klb.crosswalk_label(conn, label_org, origin_taxonomy_id=2, dst_taxonomy_id=1, always_list=True)
    assert label_dst == [('KW', 'PC')]
    assert is_equiv == True

    label_org = ("SRKW","S01")
    label_dst, is_equiv = klb.crosswalk_label(conn, label_org, origin_taxonomy_id=2, dst_taxonomy_id=1)
    assert label_dst == [('KW', 'PC')]
    assert is_equiv == False

    label_org = ("NRKW","PC")
    label_dst, is_equiv = klb.crosswalk_label(conn, label_org, origin_taxonomy_id=2, dst_taxonomy_id=3)
    assert label_dst == [('KW', 'PC')]
    assert is_equiv == False

    label_org = ("KW","PC")
    label_dst, is_equiv = klb.crosswalk_label(conn, label_org, origin_taxonomy_id=1, dst_taxonomy_id=4)
    assert label_dst == [('SRKW', 'PC')]
    assert is_equiv == True



import os
import pytest
import json
import pandas as pd
import numpy as np
import korus.db as kdb
import korus.db_util.table as ktb
from korus.util import list_to_str


current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(current_dir, "assets")
path_to_tmp = os.path.join(path_to_assets, "tmp")


if not os.path.exists(path_to_tmp):
    os.makedirs(path_to_tmp)


def test_create_db():
    """Test that we can create an empty database"""
    path = os.path.join(path_to_tmp, "empty.sqlite")
    conn = kdb.create_db(path)
    conn.close()
    assert os.path.exists(path)
    os.remove(path)

def test_get_taxonomy(basic_db):
    """Test that we can use the get_taxonomy function to retrieve taxonomies
    OBS: This test does not make any checks on the contents of the taxonomies 
    """
    (conn, path) = basic_db
    tax_latest = kdb.get_taxonomy(conn)
    assert tax_latest.version == 4
    tax_v2 = kdb.get_taxonomy(conn, taxonomy_id=2)
    assert tax_v2.version == 2

def test_get_label_id(basic_db):
    """Test that we can use get_label_id function to retrieve correct label indices"""
    (conn, path) = basic_db

    tax_id = 2

    c = conn.cursor()
    query = f"SELECT id,sound_source_tag,sound_type_tag FROM label WHERE taxonomy_id = {tax_id}"
    rows = c.execute(query).fetchall()
    df = pd.DataFrame({
        "id": [r[0] for r in rows],
        "ss": [r[1] for r in rows],
        "st": [r[2] for r in rows],
    })
    df.set_index("id", inplace=True)

    l = kdb.get_label_id(conn, ("SRKW","S01"), taxonomy_id=tax_id, ascend=False, descend=False)
    assert df.loc[l].ss == "SRKW"
    assert df.loc[l].st == "S01"

    l_list = kdb.get_label_id(conn, ("SRKW","%"), taxonomy_id=tax_id, ascend=False, descend=False)
    for l in l_list:
        assert df.loc[l].ss == "SRKW"
        assert df.loc[l].st in ["S01", "S02", "PC", "CK", "W", "TC", "Unknown"]

    l_list = kdb.get_label_id(conn, ("SRKW","S01"), taxonomy_id=tax_id, ascend=True, descend=False)
    for l in l_list:
        assert df.loc[l].ss in ["SRKW", "KW", "Mammal", "Bio", "Unknown"]
        assert df.loc[l].st in ["S01", "PC", "TC", "Unknown"]

    l_list = kdb.get_label_id(conn, ("KW","PC"), taxonomy_id=tax_id, ascend=False, descend=True)
    for l in l_list:
        assert df.loc[l].ss in ["KW", "SRKW", "NRKW"]
        assert df.loc[l].st in ["PC", "S01", "S02", "N01"]

    with pytest.raises(ValueError):
        kdb.get_label_id(conn, ("Fish","PC"), taxonomy_id=tax_id, ascend=False, descend=True)

    with pytest.raises(ValueError):
        kdb.get_label_id(conn, ("Fish","%"), taxonomy_id=tax_id, ascend=False, descend=True)

    with pytest.raises(ValueError):
        kdb.get_label_id(conn, ("%","LoudSound"), taxonomy_id=tax_id, ascend=False, descend=True)


def test_insert_deployment(basic_db, deploy_data):
    """ Test that we can add a deployment to the database"""
    (conn, path) = basic_db
    c = conn.cursor()
    c = kdb.insert_row(conn, table_name="deployment", values=deploy_data)


def test_insert_files(basic_db, deploy_data, file_data):
    """ Test that we can add audio files to the database"""
    (conn, path) = basic_db
    c = conn.cursor()
    c = kdb.insert_row(conn, table_name="deployment", values=deploy_data)

    v = {
        "name": "laptop",
        "path": "/",
        "description": "data from my latest deployments",
    }
    c = kdb.insert_row(conn, table_name="storage", values=v)

    for v in file_data:
        kdb.insert_row(conn, table_name="file", values=v)


def test_comprehensive_example(basic_db, deploy_data, file_data):
    """ Testing a fairly comprehensive example of adding annotations to the database
    TODO: refactor this test into multiple smaller tests.
    """
    (conn, path) = basic_db
    c = conn.cursor()
    c = kdb.insert_row(conn, table_name="deployment", values=deploy_data)

    v = {
        "name": "laptop",
        "path": "/",
        "description": "data from my latest deployments",
    }
    c = kdb.insert_row(conn, table_name="storage", values=v)

    for v in file_data:
        kdb.insert_row(conn, table_name="file", values=v)

    # insert job
    primary_sound = kdb.get_label_id(
        conn, 
        source_type=[("KW","PC"), ("KW","W")],
        taxonomy_id=2,
        always_list=True
    )
    background_sound = kdb.get_label_id(
        conn, 
        source_type=("%","CK"),
        taxonomy_id=2,
        always_list=True
    )
    v = {
        "taxonomy_id": 2,
        "model_id": None,
        "annotator": "LL",
        "primary_sound": json.dumps(primary_sound),
        "background_sound": json.dumps(background_sound),
        "is_exhaustive": 1,
        "configuration": None,
        "start_utc": "2022-10",
        "end_utc": "2023-03",
        "by_human": 1,
        "by_machine": 0,
        "comments": "Vessel noise annotated opportunistically",
        "issues": json.dumps(["start and end times may not always be accurate", "some KW sounds may have been incorrectly labelled as HW"]),
    }
    c = kdb.insert_row(conn, table_name="job", values=v)

    # link files
    # get deployment id
    query = """
        SELECT 
            id
        FROM 
            deployment 
        WHERE 
            owner LIKE 'OceanResearch' 
            AND name LIKE 'WestPoint'
            AND start_utc >= '2022-01-01'
            AND end_utc <= '2022-12-31'
    """
    rows = c.execute(query).fetchall()
    deploy_id = rows[0][0]

    # get file_id
    for v in file_data:
        fname = v["filename"]

        query = f"""
            SELECT 
                id
            FROM 
                file 
            WHERE 
                deployment_id = '{deploy_id}' 
                AND filename LIKE '{fname}'
        """
        rows = c.execute(query).fetchall()
        file_id = rows[0][0]

        # link file to job
        v = {
            "job_id": 1,
            "file_id": file_id,
            "channel": 0
        }
        c = kdb.insert_row(conn, table_name="file_job_relation", values=v)

    # define some tags
    v = {
        "name": "NEGATIVE",
        "description": "A negative sample"
    }
    c = kdb.insert_row(conn, table_name="tag", values=v)

    #insert annotation using add_annotation function
    annot_tbl = pd.DataFrame({
        "file_id": [1,1,1],  #id:1,2,3
        "channel": [0,0,0],
        "sound_source": ["KW","SRKW",None],
        "sound_type": ["PC","S01",None],
        "tentative_sound_source": ["SRKW",None,None],
        "tentative_sound_type": ["S01",None,None],
        "tag": [None,None,["NEGATIVE"]],
        "duration_ms": [1300,300000,800],
        "start_ms": [30000,21200,1000],
        "freq_min_hz": [600,700,None],
        "freq_max_hz": [4400,3300,None],
        "granularity": ["unit","window","window"],
        "comments": ["no additional observations","","this is a negative sample"]
    })
    annot_ids = kdb.add_annotations(conn, annot_tbl=annot_tbl, job_id=1)
    neg_ids = kdb.add_negatives(conn, job_id=1) #this adds 3 negatives

    query = "SELECT label_id,start_ms,duration_ms,file_id,tag_id FROM annotation"
    rows = c.execute(query).fetchall()
    for i,row in enumerate(rows):
        l = row[0]
        f = row[3]
        ss,st = c.execute(f"SELECT sound_source_tag,sound_type_tag FROM label WHERE id = '{l}'").fetchall()[0]
        start = row[1]/1000.
        end = start + row[2]/1000.
        tag_id = json.loads(row[4])

        if i == 0:
            assert ss == "KW"
            assert st == "PC"
            assert start == 30.0
            assert end == 31.3

        elif i == 1:
            assert ss == "SRKW"
            assert st == "S01"
            assert start == 21.2
            assert end == 321.2

        elif i == 2:
            assert ss == None
            assert st == None
            assert tag_id == [2]
            assert start == 1.0
            assert end == 1.8

        elif i > 2:
            assert ss == None
            assert st == None
            assert tag_id == [1]

            if i == 3:
                assert start == 0.0
                assert end == 1.0
            elif i == 4:
                assert start == 1.8
                assert end == 21.2
            elif i == 5:
                assert start == 21.177
                assert end == 300.0

    #test filter_annotation function
    rows = kdb.filter_annotation(conn, source_type=("KW","PC"), taxonomy_id=1)
    assert rows == [1, 2]

    rows = kdb.filter_annotation(conn, source_type=("SRKW","PC"), taxonomy_id=3)
    assert rows == [2]

    rows = kdb.filter_annotation(conn, source_type=("SRKW","S01"), taxonomy_id=2, tentative=False)
    assert rows == [2]

    rows = kdb.filter_annotation(conn, source_type=("SRKW","S01"), taxonomy_id=2, tentative=True)
    assert rows == [1, 2]

    # see all attached databases
    ##rows = conn.execute("SELECT * FROM pragma_database_list").fetchall()
    ##print(rows)

    #insert annotations with minimal required info
    annot_tbl = pd.DataFrame({
        "file_id": [2,2,2], #id:7,8,9
        "sound_source": ["KW","SRKW",None],
        "sound_type": ["PC","S01",None],
    })
    annot_ids = kdb.add_annotations(conn, annot_tbl=annot_tbl, job_id=1)

    # define some tags
    v = {
        "name": "noise",
        "description": "A sample with noise"
    }
    c = kdb.insert_row(conn, table_name="tag", values=v)
    v = {
        "name": "Loud noise",
        "description": "A sample with loud noise"
    }
    c = kdb.insert_row(conn, table_name="tag", values=v)

    #insert annotations with tags
    annot_tbl = pd.DataFrame({
        "file_id": [2,2], #id: 10,11
        "sound_source": ["SRKW","SRKW"],
        "sound_type": ["PC","PC"],
        "tag": ["noise", "Loud noise"]
    })
    annot_ids = kdb.add_annotations(conn, annot_tbl=annot_tbl, job_id=1)

    rows = kdb.filter_annotation(conn, source_type=("SRKW","PC"), taxonomy_id=3)
    assert rows == [2, 8, 10, 11]

    rows = kdb.filter_annotation(conn, source_type=("SRKW","PC"), taxonomy_id=3, tag="Loud noise")
    assert rows == [11]

    rows = kdb.filter_annotation(conn, source_type=("SRKW","PC"), taxonomy_id=3, tag="noise")
    assert rows == [10]

    rows = kdb.filter_annotation(conn, source_type=("SRKW","PC"), taxonomy_id=2, tag=["noise", "Loud noise"])
    assert rows == [10, 11]

    annot_tbl = kdb.get_annotations(conn, indices=[1,3,5])
    path = os.path.join(path_to_assets, "compr-example-test-annot1.csv")
    expected = pd.read_csv(path)
    expected = expected.astype({
        "start_utc": "datetime64[ns]", 
        "tentative_sound_source": "object",
        "tentative_sound_type": "object",
        "machine_prediction": "object",
        "ambiguous_label": "object",
    })
    #expected.ambiguous_label = expected.ambiguous_label.fillna("")
    def _decode_tag(x):
        if isinstance(x, float) and np.isnan(x):
            return None
        else:
            return json.loads(x)
    expected.tag = expected.tag.apply(lambda x: _decode_tag(x))
    pd.testing.assert_frame_equal(annot_tbl[expected.columns], expected[expected.columns])


    indices_0 = kdb.filter_annotation(conn, tag="NEGATIVE")
    indices_1 = kdb.filter_annotation(conn, tag=ktb.AUTO_NEG)
    df_0 = kdb.get_annotations(conn, indices_0, format="ketos", label=0)
    df_1 = kdb.get_annotations(conn, indices_1, format="ketos", label=1)
    df_kt = pd.concat([df_0, df_1])

    # temporary fix: reformat to match expectatin
    df_kt.reset_index(inplace=True)
    df_kt.drop(columns=["annot_id","top_path"], inplace=True)

    path = os.path.join(path_to_assets, "compr-example-test-annot2.csv")
    expected = pd.read_csv(path)
    pd.testing.assert_frame_equal(df_kt, expected)


    #insert annotations with ambiguous labels
    annot_tbl = pd.DataFrame({
        "file_id": [2, 2, 2, 2, 2],#id: 12,13,14,15,16
        "sound_source": ["SRKW", "SRKW", "NRKW", "KW", "KW"],
        "sound_type": ["S01", "S02", "N01", "PC", "PC"],
        "ambiguous_sound_source": [None, None, None, "SRKW,NRKW", None],
        "ambiguous_sound_type": [None, None, None, "S01,S02,S16,N01,N22", "PC,W"],
    })
    annot_ids = kdb.add_annotations(conn, annot_tbl=annot_tbl, job_id=1, error="ignore")
    rows = c.execute(f"SELECT label_id, ambiguous_label_id FROM annotation WHERE id IN {list_to_str(annot_ids)}").fetchall()
    assert rows[0] == (36, '[null]')
    assert rows[1] == (37, '[null]')
    assert rows[2] == (43, '[null]')
    assert rows[3] == (24, '[36, 37, 43]')
    assert rows[4] == (24, '[24, 25]')

    #filter on label_id and tentative_label_id only
    idx = kdb.filter_annotation(conn, source_type=("SRKW","S02"), tentative=True, taxonomy_id=2)
    assert len(idx) == 1

    #filter again, now also including ambiguous label assignments
    idx = kdb.filter_annotation(conn, source_type=("SRKW","S02"), tentative=True, ambiguous=True, taxonomy_id=2)
    assert len(idx) == 2

    #invert the filter
    idx = kdb.filter_annotation(conn, source_type=("SRKW","PC"), invert=True, tentative=True, taxonomy_id=2)
    assert idx[-1] == annot_ids[2]
    with pytest.raises(NotImplementedError):
        idx = kdb.filter_annotation(conn, source_type=("SRKW","PC"), invert=True, tentative=True, ambiguous=True, taxonomy_id=2)
        #assert idx[-1] == annot_ids[2]
 
    #insert a humpback annotation
    annot_tbl = pd.DataFrame({
        "file_id": [2], #id:17
        "sound_source": ["HW"],
        "sound_type": ["TC"],
    })
    annot_ids = kdb.add_annotations(conn, annot_tbl=annot_tbl, job_id=1)

    #filter using @invert=True 
    idx = kdb.filter_annotation(conn, source_type=("KW","%"), invert=True, taxonomy_id=3)

    assert idx == [3, 9, 17] 

    #filter on negatives
    idx = kdb.filter_negative(conn, source_type=("KW","%"), taxonomy_id=3)
    assert len(idx) == 0
    # since only KW,PC and KW,W where subject to systematic annotation (while KW,EC wasn't), the auto-generated 
    # negatives are excluded from this search

    idx = kdb.filter_negative(conn, source_type=("KW","PC"), taxonomy_id=3)
    assert idx == [4, 5, 6]

    idx = kdb.filter_negative(conn, source_type=("HW","%"), taxonomy_id=3)
    assert len(idx) == 0


    #insert annotations with excluded labels
    annot_tbl = pd.DataFrame({
        "file_id": [2, 2, 2], #id:18,19,20
        "sound_source": ["Unknown", "Unknown", "KW"],
        "excluded_sound_source": ["KW", ["KW","HW"], "SRKW"],
        "excluded_sound_type": ["PC", "Unknown", "S01"],
    })
    annot_ids = kdb.add_annotations(conn, annot_tbl=annot_tbl, job_id=1, error="ignore")

    # check that the first two annotations (18,19) are returned when searching for non-KW annotations
    idx = kdb.filter_annotation(conn, source_type=("KW","%"), invert=True, taxonomy_id=2)
    assert idx == [3, 9, 17, 18, 19]

    # check that all three annotations (18,19,20) are returned when searching for non-(SRKW,S01) annotations
    idx = kdb.filter_annotation(conn, source_type=("SRKW","S01"), invert=True, taxonomy_id=2)
    assert idx == [3, 9, 13, 14, 17, 18, 19, 20]

    # check that the first two annotations (18,19) are returned when searching for non-(SRKW,S02) annotations
    idx = kdb.filter_annotation(conn, source_type=("SRKW","S02"), invert=True, taxonomy_id=2)
    assert idx == [2, 3, 8, 9, 12, 14, 17, 18, 19]

    # check that second and third annotations (19,20) is returned when searching for non-HW annotations
    idx = kdb.filter_annotation(conn, source_type=("HW","%"), invert=True, taxonomy_id=2)
    assert idx == [1, 2, 3, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 19, 20]


    # check that `exclude` arg works as intended when filtering
    # expecting query to return annotations 13, 14, and 20
    idx = kdb.filter_annotation(conn, source_type=("KW","%"), exclude=("SRKW", "S01"), taxonomy_id=2)
    assert idx == [13, 14, 20]


def test_import_taxonomy(basic_db):
    """ Test import_taxonomy function"""
    (conn, src_path) = basic_db
    conn.close()

    dst_path = os.path.join(path_to_assets, "db-test-import.sqlite")
    if os.path.exists(dst_path):
        os.remove(dst_path)

    conn = kdb.create_db(dst_path)

    c = kdb.import_taxonomy(conn, src_path, name="SimpleTax", new_name="NewName")

    c = conn.cursor()
    rows = c.execute("SELECT name,version FROM taxonomy").fetchall()
    assert len(rows) == 4
    for i,r in enumerate(rows):
        assert r[0] == "NewName"
        assert r[1] == i + 1

    os.remove(dst_path)


def test_custom_column(basic_db, deploy_data):
    """ Test that we can add a custom column to a table"""
    (conn, path) = basic_db

    conn.execute("ALTER TABLE deployment ADD COLUMN cost REAL")

    deploy_data["cost"] = 30

    c = kdb.insert_row(conn, table_name="deployment", values=deploy_data)

    rows = c.execute("SELECT cost FROM deployment").fetchall()
    assert rows[0][0] == 30

    conn.commit()
